local chaosversion = "v4.1"

--[[ Sonic Robo Blast 2: Chaos Edition
Author: birbhorse, aka EeveeEuphoria
Additional credits: Mirror Mode/MirrorControls_v2.1.lua by QuickWolfPT
Additional credits: BootlegAttract by Tatsuru (found in the slot machines addon)
Additional credits: customsave_io.lua (Serialize and Deserialize functions) by sprkizard AKA szv (found here: https://github.com/sprkizard/srb2-lua-lib/blob/master/game/customsave_io.lua)

What this is: A mod that is akin to Super Mario 64 Chaos Edition, where random effects are applied to the player, generally to hinder their progress.
Every 5-12 seconds, there's a random chance a status effect will be applied to the player. Each effect has a different chance of happening, as opposed to them all having equal chances to occur.
The player has 4 possible slots for effects to occur in. If all slots are somehow filled up, the player will not have any more effects applied to them for the moment.

]]-- 

--[[ CHANGELOG:
	
	[*] Changed darkened vision, now the animation is smoother.
	[*] Added support for MMT's menu system, available in MMT v4.0+!
	[*] Slightly increased the odds of a medium sized pizza.
	[*] Added a new topping to the pizza!
	[*] Black hole no longer affects the player if they're in a minecart, or on a rollout rock.
	[*] Fixed a bug where the aggressive pizza would not play the sound on the correct target, making it playing globally instead.
	
]]

--[[ WANT TO ADD YOUR OWN EFFECTS?
This script is a complicated mess of things to get working, I know, but here's a how-to-do list:
	
	- Scroll down until you see the "statuseffects" table. This is where you can add your effect. Give it a short and brief name, and copy it to each of the different tables.
		- The first table is used to identify which one is which.
		- "effectchance" is used to determine how likely it is an effect will be applied when the effect is rolled. The higher the number in X (FRACUNIT/X) is, the less likely it'll happen.
		- And "effectduration" is how long the effect will last in seconds. If it's a one-time fx, set it to 0.
	
	- Depending on what you want, you'll either want to scroll down and find -- [ one-time fx ] -- , or -- [ timer fx, starting section ] --
		- one-time fx applies instantly the player, and immediately affects something like their ring counter, and is removed the tic after it's put in place.
		- timer fx happens over time, such as cement shoes or earthquake.
	
	- Scroll down into the PlayerThink code, and copy something like "earthquake". If you're doing a one-time fx, you can use "random music" as an example.
		- Replace all the references to quake with something else, look at these to be sure: (note, does not apply to one-time fx)
			- QuakeNow
			- QuakeRemoveNow
			- QuakeTimer
		- For either case of it being one-time fx or a timer fx, make sure you also replace any references to their original effects to the one you're making.
			- CheckContiniousEffects(player, "earthquake")
			- IsPermanentEffect("earthquake") -- this does not apply to one-time fx
			- RemoveEffect("earthquake") 
	
	- If you're doing a timer fx, go to the PlayerSpawn hook and add these to the list of pre-existing stuff that you see, and set the values to be their respective stuff. Just like this:
		- QuakeNow = false
		- QuakeRemoveNow = false
		- QuakeTimer = 0
		
	- The chaosbad variable can be checked for, if it's set to "on", generally what you want to do is make the effect worse if it's a negative effect, or even better if it's a positive effect.

	- This also only applies to timer fx stuff, look up this: local function RemoveMeNow(effect)
		- Copy and paste this to the bottom of the function (before it says end), and replace it with what you need.
		- elseif effect == "earthquake"
			player.chaos.QuakeRemoveNow = true

	- With all that out of the way, you'll want to go into the chaoshud script and get your effect nestled in there too.
		- Look for "one-time fx" and "timer fx" again in that script, you'll be copying and pasting again.
		- For the timer fx, find something that says CENUM == "earthquake" , once you do, replace the text with what you want. And then replace both instances of player.chaos.QuakeTimer with the relevant stuff you need.
		- Follow the same-thing as above for one-time fx, but don't worry about the timer.
		- And for both, you'll need to supply HUD elements for it to look good. 
			- If you don't want to worry about this right away, it is okay to use a placeholder sprite for the time being. Just be sure to add one later!
			- Look in the "Sprites" folder in the PK3, you'll find sprites labeled stuff like "CEICC0" and "CEICZ0", these are the sprites you need to work with. Be sure it starts with "CEIC", not "CEIB"!
			- As a reference for the kinds of effects, negative is "CEICC0", neutral is "CEICJ0", and positive is "CEICN0"
			- Follow the naming conventions found here in order to name your file after making your HUD sprite, be sure to put it last: http://wiki.srb2.org/wiki/Sprite#Frame_list
				- Do note that it doesn't have to be strictly "CEIC", so if you somehow run out of room to put effects in, put something like "SPR_CEID" in the freeslot table, and keep going!
			- Once all of that is done, go back into chaoshud.lua and put your file name in the ApplyEffect function. Be sure there isn't a file extension after it (i.e. png or lmp)!
	
	- Finally, search for this: local function LastCheck(player, arg1)
		- It is highly recommended you check to see if your effect would somehow conflict with either another effect, or something in the game world (i.e. minecarts, rolling balls, player being invunerable, etc.)
		- If you think your effect would have issues with that, place it in there by using: 
		if arg1 = "earthquake" and player.rings = 0 
			debugprint("earthquake not applied due to player not having rings") -- recommended to put debug print here, but not required!
			return false
		end
		
	- Once all of this is done, I recommend you use the "chaoseffectbutton" to try out your effect via a bind!
	
	- And with this huge list out of the way, congrats, you've made your effect! I know it's complicated, and one day I might simplify this whole thing down so it can be much easier to make your own effects.
]]


--[[ INCOMPLETE LIST OF THINGS BAD MODE DOES:

	- Reduce wait time in-between the time in which effects can be used again.
	- Certain effects have had their timers extended by up to 5 seconds.
	- "magnet" now affects springs.
	- "cement shoes" has increased it's amount of grip on the player. 
		- It modifies how the spin dash works too. You charge it up quicker, but you don't get much speed.
	- "the pizza is aggressive" goes through shields.
	- "negative rings" doubles the amount of rings you lose per ring collected.

]]

--[[ UPDATE PLANS: make a message that says "you'll regret that" when turning on bad mode

	- [x] !!!!!!!!!!!!!!!!!!! TURN OFF THE "BETA" VARIABLE !!!!!!!!!!!!!!!!!!!
	
	- [ ] re-do black hole's code, use blockmap search functions 
]]

--[[ Bug Fixes To-Do

	- incompatible with the modern sonic mod
	- effects continue to be applied and roulette roll after completing a special stage
		- probably doesn't happen anymore?
	- i think something is causing the player to not be able to move when spawning, if holding down a movement key
	- characters like neo sonic also modify the camera, make it so the camera permanetly stays in the positiion it's set to be in 
	- bad mode isn't synced via netvars? maybe it's the autosave/load feature

]]


--[[ Effects:
Bad effects:
	- [ ] invisible hud 
		- doesn't apply to chaos hud, unless in bad mode, in which case it'll slowly fade away 
	- [?] Black hole, *EVERYTHING* starts coming towards the player
		- look at code for A_AttractChase
	- [?] darkened vision
		- make sure it's something you can't just remove with F3 somehow
	- [ ] reverse player
		- player is reversed to look the other way instantly, via a 180 degree angle
	- [?] i was frozen today
		- [x] freezes player for a few seconds, ala RTD from sourcemod stuff
		- [x] bad mode exclusive 
		- [ ] maybe use the ice block sprites from sonic mania? scale it according to the player's scale, or height at the very least 
			- [ ] hopefully there's a way to modify the player's sprite's Y offest in real time 
		- [x] also use the SFX from sonic mania too 
	- [ ] spawn random enemy above player
		- maybe make it more generic as "spawn object"
	- [?] floor is lava for 5 seconds
		- [?] only specific textures, the most common ones specifically (i.e. the gfz grass, and not something like the bridge)
		- [x] checks for texture via the sector function stuffs 

	- [ ] texture scrolling effect
		- have textures offsets be added + 1 using a seperate variable to do so
		- when effect is resetting, have it do - 1 until it's all reset back to normal
		- keep track of this using a seperate variable, when the seperate variable gets back to 0, everything *should* be back to normal
		- singleplayer only, would be impossible to manage in multiplayer

Neutral effects:
	- [ ] repellant
		- opposite of magnet
		- uses same check for magnet to make things simple

Good effects:
Note: Generally more rare than bad effects, but not by a large degree.
	- [ ] Double rings
		- Any time a player picks up a ring, it adds two.
		- Alternatively, just multiply the player's rings by two.
		- Icon is just two rings, simple.
	- [ ] Flicky friend
		- Summons a bird flicky, uses this sound sfx_3db06
	
]]

if CODEBASE < 220 or ((CODEBASE == 220 and SUBVERSION < 5) and not (CODEBASE > 220))
	error("SRB2 Version is not detected as v2.2.5! Please update your game to use this!", 0)
end

--[[ Chaos Edition Variables & MMT Support ]]--


local LReadChaosHelp = "no"
local LRCHNagTimes = 0
local Lmobjcount = 0

local Lchaosdebugger = "off"
local Lsomeoneusingmagnet = false
local Lsomeoneusingblackhole = false

local Lchaosbad = "off" -- owo what's this?

local Lchaoshud = "bottom"
local Lchaoshudpos = 2
local Lchaosprint = "off"

local Lmaxwaittime = 12 -- default: 12
local Lminwaittime = 5 -- default: 5
local Lvibechance = (FRACUNIT/2) + (FRACUNIT/3) -- default: (FRACUNIT/2) + (FRACUNIT/3)

local Ldefaultmaxwaittime = 12
local Ldefaultminwaittime = 6
local Ldefaultvibechance = (FRACUNIT/2) + (FRACUNIT/3)

local Lchaosautoload = "on"
local Lchaosautosave = "on"

local Lpermeffect1 = "none"
local Lpermeffect2 = "none"

local LCEbutton = "off"
local LCEbuttonfx = "none"

local Lbannedeffects = {}
local Lincrementeffects = 1

rawset(_G, "ReadChaosHelp", LReadChaosHelp)
rawset(_G, "RCHNagTimes", LRCHNagTimes)
rawset(_G, "mobjcount", Lmobjcount)

rawset(_G, "chaosdebugger", Lchaosdebugger)
rawset(_G, "someoneusingmagnet", Lsomeoneusingmagnet)
rawset(_G, "someoneusingblackhole", Lsomeoneusingblackhole)

rawset(_G, "chaosbad", Lchaosbad)

rawset(_G, "chaoshud", Lchaoshud)
rawset(_G, "chaoshudpos", Lchaoshudpos)
rawset(_G, "chaosprint", Lchaosprint)

rawset(_G, "maxwaittime", Lmaxwaittime)
rawset(_G, "minwaittime", Lminwaittime)
rawset(_G, "vibechance", Lvibechance)

rawset(_G, "defaultmaxwaittime", Ldefaultmaxwaittime)
rawset(_G, "defaultminwaittime", Ldefaultminwaittime)
rawset(_G, "defaultvibechance", Ldefaultvibechance)

rawset(_G, "chaosautoload", Lchaosautoload)
rawset(_G, "chaosautosave", Lchaosautosave)

rawset(_G, "permeffect1", Lpermeffect1)
rawset(_G, "permeffect2", Lpermeffect2)

rawset(_G, "CEbutton", LCEbutton)
rawset(_G, "CEbuttonfx", LCEbuttonfx)

rawset(_G, "bannedeffects", Lbannedeffects)
rawset(_G, "incrementeffects", Lincrementeffects)

local LVLOT_ChaosEdition = {}
rawset(_G, "VLOT_ChaosEdition", LVLOT_ChaosEdition)

local Lstatuseffects = {
["giant"] = 1,
["cement shoes"] = 2,
["tiny"] = 3,
["no charability"] = 4,
["magnet"] = 5,
["fling player"] = 6,
["anti-superring"] = 7,
["superring"] = 8,
["i taketh"] = 9,
["nuke"] = 10,
["drifting left"] = 11,
["butter physics"] = 12,
["give good stuff"] = 13,
["change skin color"] = 14,
["change skin"] = 15,
["super jump"] = 16,
["the ride never ends"] = 17,
["the pizza is aggressive"] = 18,
["remove random effect"] = 19,
["money bag"] = 20,
["ghost"] = 21, 
["earthquake"] = 22, 
["gravity coil"] = 23, 
["invert controls"] = 24, 
["change sprite"] = 25, 
["funky camera"] = 26,
["random music"] = 27, 
["the floor is lava"] = 28, 
["negative rings"] = 29, 
["darkened vision"] = 30,
["no turning"] = 31,
["i was frozen today"] = 32,
["randomize sfx"] = 33,
["black hole"] = 34,
["spawn object"] = 35,
}

local Lstatuseffects2 = Lstatuseffects

local Lstatuseffects3 = {
["giant"] = 1,
["cement shoes"] = 2,
["tiny"] = 3,
["no charability"] = 4,
["magnet"] = 5,
["drifting left"] = 6,
["butter physics"] = 7,
["change skin color"] = 8,
["change skin"] = 9,
["super jump"] = 10,
["the ride never ends"] = 11,
["the pizza is aggressive"] = 12,
["money bag"] = 13,
["ghost"] = 14,
["earthquake"] = 15,
["gravity coil"] = 16,
["invert controls"] = 17,
["change sprite"] = 18,
["funky camera"] = 19,
["the floor is lava"] = 20,
["negative rings"] = 21,
["darkened vision"] = 22,
["no turning"] = 23,
["i was frozen today"] = 24,
["randomize sfx"] = 25,
["black hole"] = 26,
}

-- i am incredibly lazy
for key, value in pairs(Lstatuseffects) do
	Lstatuseffects[key] = (value + 1)
end
for key, value in pairs(Lstatuseffects3) do
	Lstatuseffects3[key] = (value + 1)
end
Lstatuseffects["none"] = 1
Lstatuseffects3["none"] = 1

local Lhudpos = {
["0"] = 1,
["1"] = 2,
["2"] = 3,
["3"] = 4,
["4"] = 5,
["5"] = 6,
}

local Lhudpos2 = {
["top"] = 1,
["bottom"] = 2,
["off"] = 3,
}

rawset(_G, "VLOT_StatusEffects", Lstatuseffects)
rawset(_G, "VLOT_StatusEffects2", Lstatuseffects2)
rawset(_G, "VLOT_StatusEffects3", Lstatuseffects3)
rawset(_G, "VLOT_HUDPos", Lhudpos)
rawset(_G, "VLOT_HUDPos2", Lhudpos2)

local AddToMMT = false 

local function AddChaosToMMT()
	if not AddToMMT and mmtVersion 
		if MMT_AddModToMMT
			MMT_AddModToMMT("Chaos Edition", "Chaotic settings for a chaotic mod!", "VLOT_ChaosEdition")
			MMT_AddModSettings("Chaos Edition", "Chaos Bad Mode", "chaosbad", "VLOT_OnOff", "Causes effects to get a lot more extreme; both good and bad effects get more intense.")
			MMT_AddModSettings("Chaos Edition", "Perm Effect 1", "permeffect1", "VLOT_StatusEffects3", "Induces a permanent effect on the player.", nil, "chaos")
			MMT_AddModSettings("Chaos Edition", "Perm Effect 2", "permeffect2", "VLOT_StatusEffects3", "Induces a second permanent effect on the player.", nil, "chaos")
			MMT_AddModSettings("Chaos Edition", "Chaos Effect Button", "CEbutton", "VLOT_SKButton", "Instantly applies an effect to the player when pressed.", nil, "chaos")
			MMT_AddModSettings("Chaos Edition", "Apply Chaos Effect", "CEbuttonfx", "VLOT_StatusEffects", "When pressing the Chaos Effect Button, this effect will be applied.", nil, "chaos")
			MMT_AddModSettings("Chaos Edition", "Chaos HUD Position", "chaoshudpos", "VLOT_HUDPos", "Changes in increments how high or low the HUD should be.", nil, "chaos")
			MMT_AddModSettings("Chaos Edition", "Chaos HUD Placement", "chaoshud", "VLOT_HUDPos2", "Places the HUD at the top or bottom, or disables the HUD entirely.", nil, "chaos")
			MMT_AddModSettings("Chaos Edition", "Chaos Print", "chaosprint", "VLOT_OnOff", "Shows what status effects are applied to the player in the console.")
			MMT_AddModSettings("Chaos Edition", "Chaos Auto Load", "chaosautoload", "VLOT_OnOff", "Automatically loads Chaos Edition settings.")
			MMT_AddModSettings("Chaos Edition", "Chaos Auto Save", "chaosautosave", "VLOT_OnOff", "Automatically saves Chaos Edition settings.")
			MMT_AddModSettings("Chaos Edition", "Chaos Debugger", "chaosdebugger", "VLOT_OnOff", "Toggles debugging prints for Chaos Edition.")
			--MMT_AddModSettings("Chaos Edition", "Banned Effects", "bannedeffects", "VLOT_StatusEffects2", "Prevents certain effects from applying to the player.", nil, nil, false, true) -- does not work?
			AddToMMT = true
		else
			print("\x82\Notice: \x80" .. "\x80 Many Minor Tweaks detected, but is outdated! Update MMT to change Chaos Edition settings within MMT!")
			AddToMMT = true
		end
	end
end

AddChaosToMMT()

addHook("ThinkFrame", function() 
	if mmtVersion AddChaosToMMT() end
end)

--minwaittime
--maxwaittime
--vibechance
--bannedeffects -- table

freeslot(
"sfx_antiri",
"MT_JIMBO",
"S_CHAOSPIZZA",
"SPR_CHPI",
"sfx_chapis",
"sfx_chapki",
"sfx_boo",
"sfx_slappy",
"sfx_chgc",
"sfx_antico",
"sfx_clok",
"S_CHAOSEXPLOSPIZZA"
)

states[S_CHAOSPIZZA] = {
	sprite = SPR_CHPI,
	tics = -1
}

states[S_CHAOSEXPLOSPIZZA] = {
	sprite = SPR_CHPI,
	frame = B,
	tics = -1
}

mobjinfo[MT_JIMBO] = {
	doomednum = -1,
	spawnstate = S_CHAOSPIZZA,
	speed = 36*FRACUNIT,
	height = 128*FRACUNIT,
	radius = 25*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP
}

sfxinfo[sfx_chapis] = {
	singular = false,
	priority = 64,
	flags = SF_X4AWAYSOUND,
	caption = "Thud-um"
}

sfxinfo[sfx_chapki] = {
	singular = false,
	priority = 129,
	caption = "\"Oh shit.\""
}

sfxinfo[sfx_boo] = {
	singular = false,
	priority = 64,
	caption = "BoOoO!"
}

sfxinfo[sfx_slappy] = {
	singular = false,
	priority = 64,
	caption = "Slapped"
}

sfxinfo[sfx_chgc] = {
	singular = false,
	priority = 64,
	caption = "Coiled spring"
}

-- [ Cloned SFX ] --
sfxinfo[sfx_clok] = {
	singular = false, 
	priority = 96, 
	caption = "Thok"
}


-- [[[ Chaos Effects ]]] --

local statuseffects = { -- "instant death", "inevitable game crash"
["giant"] = 1,
["cement shoes"] = 2,
["tiny"] = 3,
["no charability"] = 4,
["magnet"] = 5,
["fling player"] = 6,
["anti-superring"] = 7,
["superring"] = 8,
["i taketh"] = 9,
["nuke"] = 10,
["drifting left"] = 11,
["butter physics"] = 12,
["give good stuff"] = 13,
["change skin color"] = 14,
["change skin"] = 15,
["super jump"] = 16,
["the ride never ends"] = 17,
["the pizza is aggressive"] = 18,
["remove random effect"] = 19,
["money bag"] = 20,
["ghost"] = 21, 
["earthquake"] = 22, 
["gravity coil"] = 23, 
["invert controls"] = 24, 
["change sprite"] = 25, 
["funky camera"] = 26,
["random music"] = 27, 
["the floor is lava"] = 28, 
["negative rings"] = 29, 
["darkened vision"] = 30,
["no turning"] = 31,
["i was frozen today"] = 32,
["randomize sfx"] = 33,
["black hole"] = 34,
["spawn object"] = 35,
}

local effectchance = {
["giant"] = FRACUNIT/9,
["cement shoes"] = FRACUNIT/9,
["tiny"] = FRACUNIT/6,
["no charability"] = FRACUNIT/12,
["magnet"] = FRACUNIT/5,
["fling player"] = FRACUNIT/8,
["anti-superring"] = FRACUNIT/14,
["superring"] = FRACUNIT/15,
["i taketh"] = FRACUNIT/25,
["nuke"] = FRACUNIT/50,
["drifting left"] = FRACUNIT/25,
["butter physics"] = FRACUNIT/10,
["give good stuff"] = FRACUNIT/32,
["change skin color"] = FRACUNIT/10,
["change skin"] = FRACUNIT/17,
["super jump"] = FRACUNIT/30,
["the ride never ends"] = FRACUNIT/24,
["the pizza is aggressive"] = FRACUNIT/37,
["remove random effect"] = FRACUNIT/22,
["money bag"] = FRACUNIT/34,
["ghost"] = FRACUNIT/24,
["earthquake"] = FRACUNIT/18,
["gravity coil"] = FRACUNIT/20,
["invert controls"] = FRACUNIT/29,
["change sprite"] = FRACUNIT/10,
["funky camera"] = FRACUNIT/23,
["random music"] = FRACUNIT/14,
["the floor is lava"] = FRACUNIT/35,
["negative rings"] = FRACUNIT/16,
["darkened vision"] = FRACUNIT/23,
["no turning"] = FRACUNIT/35,
["i was frozen today"] = FRACUNIT/25,
["randomize sfx"] = FRACUNIT/9,
["black hole"] = FRACUNIT/17,
["spawn object"] = FRACUNIT/18,
}

local effectduration = {
["giant"] = 10,
["cement shoes"] = 8,
["tiny"] = 8,
["no charability"] = 15,
["magnet"] = 15,
["fling player"] = 0,
["anti-superring"] = 0,
["superring"] = 0,
["i taketh"] = 0,
["nuke"] = 0,
["drifting left"] = 13, -- 10 in not bad mode
["butter physics"] = 18, -- 15 in not bad mode
["give good stuff"] = 0,
["change skin color"] = 3,
["change skin"] = 4,
["super jump"] = 12,
["the ride never ends"] = 18, -- 15 in not bad mode 
["the pizza is aggressive"] = 13, -- 10 in not bad mode
["remove random effect"] = 0,
["money bag"] = 4,
["ghost"] = 8,
["earthquake"] = 13,
["gravity coil"] = 15,
["invert controls"] = 14,
["change sprite"] = 18,
["funky camera"] = 10,
["random music"] = 0,
["the floor is lava"] = 10, -- 5 in not bad mode
["negative rings"] = 15, -- 11 in not bad mode
["darkened vision"] = 15,
["no turning"] = 12,
["i was frozen today"] = 5,
["randomize sfx"] = 20,
["black hole"] = 15,
["spawn object"] = 0,
}

rawset(_G, "chaosEffectDuration", effectduration)


-- [[[ Functions ]]] --

--maxwaittime, minwaittime, vibechance

-- Converts a Lua table to a text representation.
local function Serialize(input)
	if type(input) == "string" then
		return "[[" .. input:gsub("](%--)]", "]-%1]") .. "]]"
	elseif type(input) == "number" then
		return "[" .. input .. "]"
	elseif type(input) == "boolean" then
		if input then
			return "T"
		else
			return "F"
		end
	elseif type(input) == "nil" then
		return "N"
	elseif type(input) == "table" then

		local str = "{"
		
		local delimit = false
		for k,v in pairs(input) do
			if delimit then
				str = $ .. ";"
			end
			
			str = $ .. Serialize(k) .. ":" .. Serialize(v)
			
			delimit = true
		end
		
		str = $ .. "}"
		return str
		
	else
		error("table contains a non-serializable element!")
	end
end

-- Converts the text representation of a Lua table back to the original table.
local function Deserialize(input)
	local output

	if input:sub(1, 2) == "[[" then -- string
		local endpos = input:find("]]")
		output = input:sub(3, endpos-1):gsub("]%-(%--)]", "]%1]")
		input = $:sub(endpos+2)
	elseif input:sub(1, 1) == "[" then -- number
		local endpos = input:find("]")
		output = tonumber(input:sub(2, endpos-1))
		input = $:sub(endpos+1)
	elseif input:sub(1, 1) == "T" then -- true
		output = true
		input = $:sub(2)
	elseif input:sub(1, 1) == "F" then -- false
		output = false
		input = $:sub(2)
	elseif input:sub(1, 1) == "N" then -- nil
		output = nil
		input = $:sub(2)
	elseif input:sub(1, 2) == "{}" then -- empty table
		output = {}
		input = $:sub(3)
	elseif input:sub(1, 1) == "{" then -- table
		output = {}
		
		while input:sub(1, 1) ~= "}" do
			local key, value
			
			key, input = Deserialize(input:sub(2))
			if input:sub(1, 1) ~= ":" then
				error("string is badly formatted! "..input)
			end
			
			value, input = Deserialize(input:sub(2))
			
			output[key] = value
		end
		
		input = $:sub(2)
	else
		error("string is badly formatted! "..input)
	end
	
	if input:len() then
		return output, input
	else
		return output
	end
end

local function tablelength(T)
  local count = 0
  for _ in pairs(T) do count = count + 1 end
  return count
end

local function tableempty(self) --taken from stack overflow
    for _, _ in pairs(self) do
        return false
    end
    return true
end

local function CH_SaveSettings(player, silent) -- this is sloppy too i'm so sorry
	local CHchaoshud = io.openlocal("chaosedition/chaoshud.txt", "w")
		CHchaoshud:write(chaoshud)
		CHchaoshud:close()	
	local CHCEbutton = io.openlocal("chaosedition/CEbutton.txt", "w")
		CHCEbutton:write(CEbutton)
		CHCEbutton:close()	
	local CHCEbuttonfx = io.openlocal("chaosedition/CEbuttonfx.txt", "w")
		CHCEbuttonfx:write(CEbuttonfx)
		CHCEbuttonfx:close()
	local CHchaosdebugger = io.openlocal("chaosedition/chaosdebugger.txt", "w")
		CHchaosdebugger:write(chaosdebugger)
		CHchaosdebugger:close()
	local CHchaosautoload = io.openlocal("chaosedition/chaosautoload.txt", "w")
		CHchaosautoload:write(chaosautoload)
		CHchaosautoload:close()
	local CHchaosautosave = io.openlocal("chaosedition/chaosautosave.txt", "w")
		CHchaosautosave:write(chaosautosave)
		CHchaosautosave:close()	
	local CHchaoshudpos = io.openlocal("chaosedition/chaoshudpos.txt", "w")
		CHchaoshudpos:write(chaoshudpos)
		CHchaoshudpos:close()	
	
	local CHminwaittime = io.openlocal("chaosedition/minwaittime.txt", "w")
		CHminwaittime:write(minwaittime)
		CHminwaittime:close()
	local CHmaxwaittime = io.openlocal("chaosedition/maxwaittime.txt", "w")
		CHmaxwaittime:write(maxwaittime)
		CHmaxwaittime:close()	
	local CHvibechance = io.openlocal("chaosedition/vibechance.txt", "w")
		CHvibechance:write(vibechance)
		CHvibechance:close()	
		
	local CHchaosprint = io.openlocal("chaosedition/chaosprint.txt", "w")
		CHchaosprint:write(chaosprint)
		CHchaosprint:close()

	local CHReadChaosHelp = io.openlocal("chaosedition/ReadChaosHelp.txt", "w")
		CHReadChaosHelp:write(ReadChaosHelp)
		CHReadChaosHelp:close()	
	local CHRCHNagTimes = io.openlocal("chaosedition/RCHNagTimes.txt", "w")
		CHRCHNagTimes:write(RCHNagTimes)
		CHRCHNagTimes:close()	
		
	local CHchaosbadmode = io.openlocal("chaosedition/chaosbadmode.txt", "w")
		CHchaosbadmode:write(chaosbad)
		CHchaosbadmode:close()		

	local CHchaosbaneffect = io.openlocal("chaosedition/bannedeffects.txt", "w")
		CHchaosbaneffect:write(Serialize(bannedeffects))
		CHchaosbaneffect:close()
		
	if player CONS_Printf(player, "\x81(Chaos Edition)\x8E Saved\x80 files to: \x82" .. "/chaosedition/") end
	if silent != true and player S_StartSound(player.mo, sfx_addfil) end
end

rawset(_G, "CH_SaveSettings", CH_SaveSettings)

local function CH_LoadSettings()
	local CHchaoshud = io.openlocal("chaosedition/chaoshud.txt", "r")
		if CHchaoshud chaoshud = CHchaoshud:read("*all") CHchaoshud:close() end	
	local CHCEbutton = io.openlocal("chaosedition/CEbutton.txt", "r")
		if CHCEbutton CEbutton = CHCEbutton:read("*all") CHCEbutton:close() end	
	local CHCEbuttonfx = io.openlocal("chaosedition/CEbuttonfx.txt", "r")
		if CHCEbuttonfx CEbuttonfx = CHCEbuttonfx:read("*all") CHCEbuttonfx:close() end
	local CHchaosdebugger = io.openlocal("chaosedition/chaosdebugger.txt", "r")
		if CHchaosdebugger chaosdebugger = CHchaosdebugger:read("*all") CHchaosdebugger:close() end
	local CHchaosautoload = io.openlocal("chaosedition/chaosautoload.txt", "r")
		if CHchaosautoload chaosautoload = CHchaosautoload:read("*all") CHchaosautoload:close() end	
	local CHchaosautosave = io.openlocal("chaosedition/chaosautosave.txt", "r")
		if CHchaosautosave chaosautosave = CHchaosautosave:read("*all") CHchaosautosave:close() end
	local CHchaoshudpos = io.openlocal("chaosedition/chaoshudpos.txt", "r")
		if CHchaoshudpos chaoshudpos = tonumber(CHchaoshudpos:read("*all")) CHchaoshudpos:close() end		
	
	local CHminwaittime = io.openlocal("chaosedition/minwaittime.txt", "r")
		if CHminwaittime minwaittime = tonumber(CHminwaittime:read("*all")) CHminwaittime:close() end	
	local CHmaxwaittime = io.openlocal("chaosedition/maxwaittime.txt", "r")
		if CHmaxwaittime maxwaittime = tonumber(CHmaxwaittime:read("*all")) CHmaxwaittime:close() end	
	local CHvibechance = io.openlocal("chaosedition/vibechance.txt", "r")
		if CHvibechance vibechance = tonumber(CHvibechance:read("*all")) CHvibechance:close() end	
	
	local CHchaosprint = io.openlocal("chaosedition/chaosprint.txt", "r")
		if CHchaosprint chaosprint = CHchaosprint:read("*all") CHchaosprint:close() end	
	
	local CHReadChaosHelp = io.openlocal("chaosedition/ReadChaosHelp.txt", "r")
		if CHReadChaosHelp ReadChaosHelp = CHReadChaosHelp:read("*all") CHReadChaosHelp:close() end	
	local CHRCHNagTimes = io.openlocal("chaosedition/RCHNagTimes.txt", "r")
		if CHRCHNagTimes RCHNagTimes = tonumber(CHRCHNagTimes:read("*all")) CHRCHNagTimes:close() end
		
	local CHchaosbadmode = io.openlocal("chaosedition/chaosbadmode.txt", "r")
		if CHchaosbadmode chaosbad = CHchaosbadmode:read("*all") CHchaosbadmode:close() end		
	
	local CHchaosbaneffect = io.openlocal("chaosedition/bannedeffects.txt", "r")
		if CHchaosbaneffect bannedeffects = Deserialize(CHchaosbaneffect:read("*all")) CHchaosbaneffect:close() end	
	incrementeffects = tablelength(bannedeffects) + 1
end

rawset(_G, "CH_LoadSettings", CH_LoadSettings)

local numberofeffects = tablelength(statuseffects)

local function BootlegAttract(mo, target, speed, xoffs, yoffs, zoffs) -- originated from tatsuru's conversion of SRB2 source code
	xoffs = $ or 0
	yoffs = $ or 0
	zoffs = $ or 0
	
	local newx = target.x + xoffs
	local newy = target.y + yoffs
	local newz = target.z + zoffs
	local dist = R_PointToDist2(0, 0, R_PointToDist2(mo.x, mo.y, newx, newy), newz - mo.z)
	
	if speed < dist
		mo.momx = FixedMul(FixedDiv(newx - mo.x, dist), speed)
		mo.momy = FixedMul(FixedDiv(newy - mo.y, dist), speed)
		mo.momz = FixedMul(FixedDiv(newz - mo.z, dist), speed)
	elseif speed > dist
		mo.momx = 0
		mo.momy = 0
		mo.momz = 0
		P_TeleportMove(mo, newx, newy, newz)
	end
end

local function darkvisionhud(v, player)
	if v.TransBack == nil
		v.TransBack = 10
	end
	
	if player.chaos.DarkenedVisionNow
		if (leveltime % 2 == 0) and ((chaosbad == "on" and v.TransBack > 2) or (chaosbad == "off" and v.TransBack > 3))
			v.TransBack = $ - 1
		end
	else
		if (leveltime % 2 == 0) and v.TransBack < 10
			v.TransBack = $ + 1
		end
	end
	v.TransFlags = v.TransBack<<V_ALPHASHIFT
	v.drawScaled(0,200 , FRACUNIT*800, v.cachePatch("CEIBD0"), v.TransFlags|V_SNAPTOLEFT|V_SNAPTOTOP)
end

local function KeyToNum(key) -- taken from MMT, because it's my code bich i can do what i want with it 
	if type(key) == "string" and _G["BT_"+string.upper(key)] and not (string.upper(key) == "WEAPONMASK") --or (string.upper(key) == "JUMP")
		return _G["BT_"+string.upper(key)]
	elseif key == "spin"
		return BT_USE
	elseif key == "ringtoss"
		return BT_ATTACK
	elseif key == "ringtossnormal"
		return BT_FIRENORMAL
	else
		return nil
	end
end

local function MusicRand(num) -- ctrack = current track 
	if num == "RANGE"
		return 70
	end
	
	if num == 1
		return "TUTORL"
	elseif num == 2
		return "GFZ1"
	elseif num == 3
		return "GFZ2"
	elseif num == 4
		return "VSBOSS"
	elseif num == 5
		return "VSALT"
	elseif num == 6
		return "THZ1"
	elseif num == 7
		return "THZ2"
	elseif num == 8
		return "DSZ1"
	elseif num == 9
		return "DSZ2"
	elseif num == 10
		return "CEZ1"
	elseif num == 11
		return "CEZ2"
	elseif num == 12
		return "ACZ1"
	elseif num == 13
		return "ACZ2"
	elseif num == 14
		return "VSFANG"
	elseif num == 15
		return "RVZ1"
	elseif num == 16
		return "RVZ2"
	elseif num == 17
		return "RVZALT"
	elseif num == 18
		return "DCZ"
	elseif num == 19
		return "ERZ1"
	elseif num == 20
		return "ERZ2"
	elseif num == 21
		return "DISCO"
	elseif num == 22
		return "BCZ1"
	elseif num == 23
		return "VSMETL"
	elseif num == 24
		return "VSBRAK"
	elseif num == 25
		return "FHZ"
	elseif num == 26
		return "MARIO1"
	elseif num == 27
		return "MARIO2"
	elseif num == 28
		return "THZALT"
	elseif num == 29
		return "HAUNTD"
	elseif num == 30
		return "AGZ"
	elseif num == 31
		return "ATZ"
	elseif num == 32
		return "CCZ"
	elseif num == 33
		return "DHZ"
	elseif num == 34
		return "APZ1"
	elseif num == 35
		return "APZ2"	
	elseif num == 36
		return "SPEC1"
	elseif num == 37
		return "SPEC2"
	elseif num == 38
		return "SPEC3"
	elseif num == 39
		return "SPEC4"
	elseif num == 40
		return "SPEC5"
	elseif num == 41
		return "SPEC6"
	elseif num == 42
		return "SPEC7"
	elseif num == 43
		return "BHZ"
	elseif num == 44
		return "SPEC8"
	elseif num == 45
		return "OSPEC"
	elseif num == 46
		return "OSPEC2"
	elseif num == 47
		return "GFZALT"
	elseif num == 48
		return "AGZALT"
	elseif num == 49
		return "VSAGZ"
	elseif num == 50
		return "MP_WTR"
	elseif num == 51
		return "MP_ICE"
	elseif num == 52
		return "MP_RAI"
	elseif num == 53
		return "MP_AIR"
	elseif num == 54
		return "MP_SUN"
	elseif num == 55
		return "MP_DES"
	elseif num == 56
		return "MP_DOM"
	elseif num == 57
		return "MP_GHZ"
	elseif num == 58
		return "MP_MED"
	elseif num == 59
		return "MP_0RI"
	elseif num == 60
		return "MP_COA"
	elseif num == 61
		return "MP_FIR"
	elseif num == 62
		return "MP_FOR"
	elseif num == 63
		return "MP_FR2"
	elseif num == 64
		return "MP_GL2"
	elseif num == 65
		return "MP_GLZ"
	elseif num == 66
		return "MP_MIN"
	elseif num == 67
		return "MP_MRZ"
	elseif num == 68
		return "MP_WT2"
	elseif num == 69 -- nice
		return "_CHSEL"
	elseif num == 70
		return "_INTER"

	end
end

local function SoundRand(num) -- TODO: CLONE SFX SO IT DOESN'T CONFLICT WITH THE EXISTING SOUNDS THE THING IS CHECKING FOR
	local num = P_RandomRange(1, 18) -- last number is max limit for sounds
	
	if num == 1
		return sfx_clok -- thok
	elseif num == 2
		return sfx_antiri
	elseif num == 3
		return sfx_s3k38 -- gasp
	elseif num == 4
		return sfx_s3kba -- putput
	elseif num == 5
		return sfx_s23c -- spin
	elseif num == 6
		return sfx_s3k36 -- skid
	elseif num == 7
		return sfx_alarm
	elseif num == 8
		return sfx_crumbl
	elseif num == 9
		return sfx_laser
	elseif num == 10
		return sfx_s3k86
	elseif num == 11
		return sfx_doorc2
	elseif num == 12
		return sfx_bkpoof
	elseif num == 13
		return sfx_pogo
	elseif num == 14
		return sfx_shldls
	elseif num == 15
		return sfx_bsnipe
	elseif num == 16
		return sfx_s3k4e
	elseif num == 17
		return sfx_s3k49
	elseif num == 18
		return sfx_s3k62

	end
end

local function P_TelekinesisButBetter(player, thrust, range) -- dunno how to target players :(

	if player.powers[pw_super]
		thrust = $ * 2
	end

	local nonenemies = false
	
	if chaosbad == "on"
		nonenemies = true
	end

	local mo2 = P_LookForEnemies(player, nonenemies, true)
	
	--local mo3 = P_LookForPlayers(player.mo, range, false)
	
	
	if mo2
		local angle = R_PointToAngle2(player.mo.x, player.mo.y, mo2.x, mo2.y)
		
		if (mo2.health > 0)
			P_Thrust(mo2, angle, thrust)

			if (mo2.type == MT_GOLDBUZZ or mo2.type == MT_REDBUZZ)
				mo2.tics = 8
			end
		end
	end	
	
	--[[if mo3 and player.mo.target != player.mo
		local angle = R_PointToAngle2(player.mo.x, player.mo.y, player.mo.target.x, player.mo.target.y)
		
		if (player.mo.target.health > 0)
			P_Thrust(player.mo.target, angle, thrust)
		end
	end]]
	
	
	P_SpawnSpinMobj(player, MT_THOK)
	-- no PF_THOKKED flag for you
end

local function numberinserver() -- thanks to "Monster ɪestyn" on the SRB2 discord for this
	local c = 0
	for p in players.iterate do c = $ + 1 end
	return c
end


local function numberofskins() -- modified on code from Monster ɪestyn
	local c = 0
	for p in skins.iterate do c = $ + 1 end
	return (c-1)
end

local function tablesearch(tbl, val)
    for k, v in pairs(tbl) do
        if v == val then 
			return k 
		end
    end
    return nil
end

local function invtablesearch(tbl, val)
    for k, v in pairs(tbl) do
        if k == val then 
			return v
		end
    end
    return nil
end

local function cmdDesc(tbl, key)
    for k, v in pairs(tbl) do
        if k == key then 
			return v
		end
    end
    return nil
end

local function debugprint(text)
	if chaosdebugger == "on"
		print(text)
	end
end

local function debugconsprintf(player, text)
	if chaosdebugger == "on"
		CONS_Printf(player, text)
	end
end

local function checkOnLean(onoff, togglevar)
	if onoff == "on" or onoff == "off"
		return onoff
	elseif onoff == "toggle"
		if togglevar == "on"
			return "off"
		else
			return "on"
		end
	else
		return false
	end
end

local function changedvar(player, cmdname, var)
	CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" +var)
end

	
local function EffectPrint(player, text)
	if chaosprint == "on"
		CONS_Printf(player, text)
	end
end

local function ApplyEffect(player, arg1)
	if player.chaos.ContinuousEffect1 == nil
		player.chaos.ContinuousEffect1 = arg1
		debugconsprintf(player, "slot 1")
	elseif player.chaos.ContinuousEffect2 == nil
		player.chaos.ContinuousEffect2 = arg1
		debugconsprintf(player, "slot 2")
	elseif player.chaos.ContinuousEffect3 == nil
		player.chaos.ContinuousEffect3 = arg1
		debugconsprintf(player, "slot 3")
	elseif player.chaos.ContinuousEffect4 == nil
		player.chaos.ContinuousEffect4 = arg1
		debugconsprintf(player, "slot 4")
	end
	
	player.chaos.HUDTEXT = "RESET"
	
	if arg1 == "fling player" -- only reason why i'm doing this janky work around is so that "fling player" and "slapped" doesn't happen one after another, and i want the names to be distinct 
		local RANDYSLAP 
		
		if P_IsObjectOnGround(player.mo)
			if chaosbad == "on"
				RANDYSLAP = P_RandomChance(FRACUNIT/2 + FRACUNIT/6) -- 66.6%
			else
				RANDYSLAP = P_RandomChance(FRACUNIT/3) -- 33.3%
			end
		else
			if chaosbad == "on"
				RANDYSLAP = P_RandomChance(FRACUNIT/2 + FRACUNIT/3) -- 83.3%
			else
				RANDYSLAP = P_RandomChance(FRACUNIT/2 + FRACUNIT/6) -- 66.6%
			end
		end
		
		if RANDYSLAP
			player.chaos.Slapped = true
			player.chaos.PrintOff = "slapped"
		else
			player.chaos.PrintOff = arg1
		end
	else
		player.chaos.PrintOff = arg1
	end
	
	EffectPrint(player, "*\x83\ applied\x80: " .. player.chaos.PrintOff)
	player.chaos.HUDTEXT = player.chaos.PrintOff
	player.chaos.HUDTEXTTIMER = 0
	player.chaos.StatusEffect = nil
	player.chaos.PrintOff = nil 
	
end

local function SlotsFull(player)
	if player.chaos.ContinuousEffect1 and player.chaos.ContinuousEffect2 and player.chaos.ContinuousEffect3 and player.chaos.ContinuousEffect4
		return true
	else
		return false
	end
end

local function CheckContiniousEffects(player, arg1)
	if player.chaos.ContinuousEffect1 == arg1
		return player.chaos.ContinuousEffect1
	elseif player.chaos.ContinuousEffect2 == arg1
		return player.chaos.ContinuousEffect2
	elseif player.chaos.ContinuousEffect3 == arg1
		return player.chaos.ContinuousEffect3
	elseif player.chaos.ContinuousEffect4 == arg1
		return player.chaos.ContinuousEffect4
	elseif player.chaos.permeffect1 == arg1
		return "perm"
	elseif player.chaos.permeffect2 == arg1
		return "perm"
	else
		return false
	end
end

local function LastCheck(player, arg1) --checks for exceptions; e.g. if an effect has been previously used, or the player is in a minecart, etc. 
	if arg1 == "remove random effect" and not (player.chaos.ContinuousEffect1 or player.chaos.ContinuousEffect2 or player.chaos.ContinuousEffect3 or player.chaos.ContinuousEffect4)
		debugprint("did not choose due to no effects being present")
		return false
	end
	
	if arg1 == "the floor is lava" and (player.powers[pw_shield] & SH_PROTECTFIRE)
		debugprint("did not choose due to fire shield")
		return false
	end		
	
	if (arg1 == "the floor is lava" or arg1 == "the pizza is aggressive") and (player.powers[pw_invulnerability] and (player.powers[pw_invulnerability] > 2*TICRATE))
		debugprint("did not choose due to invincibility")
		return false
	end
	
	if arg1 == "turning left" and CheckContiniousEffects(player, "no turning")
		debugprint("did not choose due to no turning")
		return false
	end		

	if arg1 == "no turning" and CheckContiniousEffects(player, "turning left")
		debugprint("did not choose due to turning left")
		return false
	end
	
	if arg1 == "randomize sfx" and netgame
		debugprint("did not choose due to being in a netgame")
		return false
	end
	
	if arg1 == "earthquake" and numberinserver() > 2
		debugprint("did not choose because there are more than 2 people in the server")
		return false
	end
	
	if someoneusingmagnet and (arg1 == "magnet")
		debugprint("did not choose because someone else is using magnet")
		return false
	end		
	
	if arg1 == "black hole" and CheckContiniousEffects(player, "magnet")
		debugprint("did not choose because magnet is already applied")
		return false
	end		
	
	if arg1 == "magnet" and CheckContiniousEffects(player, "black hole")
		debugprint("did not choose because black hole is already applied")
		return false
	end
	
	if player.chaos.PreviousEffect1 == arg1 or player.chaos.PreviousEffect2 == arg1 or player.chaos.PreviousEffect3 == arg1 or player.chaos.PreviousEffect4 == arg1
		debugprint("did not choose due to being in a previous effect")
		return false
	end
	
	if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) and (arg1 == "the pizza is aggressive" or arg1 == "i was frozen today" or arg1 == "spawn object" or arg1 == "the ride never ends" or arg1 == "the floor is lava" or arg1 == "super jump" or arg1 == "butter physics" or arg1 == "giant" or arg1 == "magnet" or arg1 == "gravity coil")
		debugprint("did not choose due to being in NiGHTs")
		return false
	end		
	
	if (G_IsSpecialStage(gamemap)) and (arg1 == "the pizza is aggressive" or arg1 == "giant" or arg1 == "magnet" or arg1 == "the floor is lava")
		debugprint("did not choose due to being in a special stage")
		return false
	end
	
	if player.powers[pw_carry] == CR_MINECART and (arg1 == "blackhole" or arg1 == "giant" or arg1 == "the floor is lava" or arg1 == "tiny" or arg1 == "super jump" or arg1 == "cement shoes" or arg1 == "butter physics" or arg1 == "fling player" or arg1 == "gravity coil" or arg1 == "turning left")
		debugprint("did not choose due to being in a minecart")
		return false
	end
	
	if player.powers[pw_carry] == CR_ROLLOUT and (arg1 == "giant" or arg1 == "tiny" or arg1 == "butter physics" or arg1 == "fling player" or arg1 == "black hole")
		debugprint("did not choose due to being on a rollout rock")
		return false
	end
	
	if (mapheaderinfo[gamemap].bonustype == 1 or mapheaderinfo[gamemap].bonustype == 2) and (arg1 == "magnet" or arg1 == "the pizza is aggressive" or arg1 == "anti-superring" or arg1 == "negative rings")
		debugprint("did not choose due to being in boss stage")
		return false
	end
	
	if (arg1 == "cement shoes" or arg1 == "i was frozen today") and (CheckContiniousEffects(player, "no charability") or CheckContiniousEffects(player, "the pizza is aggressive") or CheckContiniousEffects(player, "change skin color"))
		debugprint("did not choose due to no charability, pizza, change skin color, no charability")
		return false
	end			
	
	if arg1 == "cement shoes" and CheckContiniousEffects(player, "the floor is lava")
		debugprint("did not choose due to the floor is lava")
		return false
	end			
	
	if arg1 == "the floor is lava" and chaosbad == "off" and CheckContiniousEffects(player, "cement shoes")
		debugprint("did not choose due to cement shoes, and not in bad mode")
		return false
	end		
	
	if (arg1 == "no charability" or arg1 == "change skin color" or arg1 == "i was frozen today") and CheckContiniousEffects(player, "cement shoes")
		debugprint("did not choose due to cement shoes")
		return false
	end	
	
	if arg1 == "the floor is lava" and player.rings == 0
		debugprint("did not choose because player has no rings")
		return false
	end
	
	if arg1 == "tiny" and CheckContiniousEffects(player, "giant")
		debugprint("did not choose due to giant")
		return false
	end
	
	if arg1 == "giant" and CheckContiniousEffects(player, "tiny")
		debugprint("did not choose due to tiny")
		return false
	end

	if chaosbad == "off" and (arg1 == "inevitable game crash" or arg1 == "i was frozen today" or arg1 == "invert controls")
		debugprint("did not choose due to chaosbad not being enabled")
		return false
	end
	
	if (arg1 == "invert controls" or arg1 == "change skin color" or arg1 == "no charability") and CheckContiniousEffects(player, "i was frozen today")
		debugprint("did not choose due to being frozen today")
		return false
	end
	
	if arg1 == "i was frozen today" and CheckContiniousEffects(player, "invert controls")
		debugprint("did not choose due to inverted controls")
		return false
	end
	
	if multiplayer and (arg1 == "inevitable game crash")
		debugprint("did not choose due to multiplayer")
		return false
	end
	
	if tablesearch(bannedeffects, arg1)
		debugprint("effect is on banned effects list, did not choose")
		return false
	end
	
	return true
end

-- [[[ Commands ]]] --

local commandDescription = {
["chaosdebugger"] = "*\x85 chaosdebugger \x81<on/off>\x80: Enables debugging printing.",
["chaoshud"] = "*\x85 chaoshud \x81<top/bottom/off>\x80, \x81<0-5>\x80: Changes the position of the HUD, or disables it.",
["chaosprint"] = "*\x85 chaosprint \x81<on/off>\x80: Toggles the printing of effects to console.",
["chaosloadsettings"] = "*\x85 chaosloadsettings\x80: Load in settings from files in a folder located in luafiles/chaosedition.",
["chaossavesettings"] = "*\x85 chaossavesettings\x80: Saves settings to files in a folder located in luafiles/chaosedition.",
["chaosautoload"] = "*\x85 chaosautoload \x81<on/off>\x80: Automatically loads the user-saved settings when loading Chaos Edition.",
["chaosautosave"] = "*\x85 chaosautosave \x81<on/off>\x80: Automatically saves Chaos Edition settings every time you change settings.",
["chaoschances"] = "*\x85 chaoschances \x81<min time (>3, <9), max time (>4, <20), chance (normal/always/rarer)>\x80: Changes the minimum/maximum time between effects, alongside the chance an effect will occur. You can set 'default' to each value.",
["chaoshelp"] = "*\x85 chaoshelp \x81<num>\x80: Lists all the available commands, by using the first argument as a page number.",
["chaospermeffects"] = "*\x85 chaospermeffects \x81\"<effect 1>\", \"<effect 2>\"\x80: Applies 1 or 2 permanent effects to the player.",
["chaosbadmode"] = "*\x85 chaosbadmode \x81<on/off>\x80: Enables Bad Mode, making everything so much worse. It also enables new effects that are really awful.",
["chaosbaneffect"] = "*\x85 chaosbaneffect \x81\"<effect>\" <remove>\x80: Disables (an) effect(s) from being applied to the player. Specifying 'remove' after the effect will remove it from the list.",
["chaoseffectbutton"] = "*\x85 chaoseffectbutton \x81<key> \"<effect>\"\x80: Immeditately applies an effect to the player, and if specified, applies a specific effect to the player."
}

local function ChaosDebugger(player, arg1)
local cmdname = "chaosdebugger"
	if netgame and not IsPlayerAdmin(player)
		CONS_Printf(player, "\x85\ERROR:\x80 Can not use while in a netgame and not an admin!")
		return
	end

	if (checkOnLean(arg1) or arg1 == "disable") and not multiplayer
		arg1 = checkOnLean(arg1, chaosdebugger)
		chaosdebugger = arg1
		changedvar(player, cmdname, arg1)
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		changedvar(player, cmdname, chaosdebugger) 
	end
end

local function ChaosAutoload(player, arg1)
local cmdname = "chaosautoload"
	if checkOnLean(arg1) and not multiplayer
		arg1 = checkOnLean(arg1, chaosautoload)
		chaosautoload = arg1
		changedvar(player, cmdname, arg1)
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		changedvar(player, cmdname, chaosautoload) 
	end
end

local function ChaosAutosave(player, arg1)
local cmdname = "chaosautosave"
	if checkOnLean(arg1) and not multiplayer
		arg1 = checkOnLean(arg1, chaosautosave)
		chaosautosave = arg1
		changedvar(player, cmdname, arg1)
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		changedvar(player, cmdname, chaosautosave) 
	end
end

local function ChaosHud(player, arg1, arg2)
local cmdname = "chaoshud"
	if tonumber(arg1) or arg1 == "0"
		arg1 = tonumber(arg1) 
	end
	if tonumber(arg2) or arg2 == "0"
		arg2 = tonumber(arg2) 
	end

	if ((arg1 == "bottom" or arg1 == "top" or arg1 == "off") and (not arg2 or (type(arg2) == "number"))) or (not arg2 and (type(arg1) == "number"))
		
		if (type(arg1) == "number") and (arg1 >= 0 and arg1 <= 5)
			arg1 = tonumber(arg1)
			if not multiplayer
				chaoshudpos = arg1
			end
			player.chaos.chaoshudpos = arg1
		else
			if not multiplayer
				chaoshud = arg1
			end
			player.chaos.chaoshud = arg1
			if (type(arg2) == "number") and (arg2 >= 0 and arg2 <= 5)
				arg2 = tonumber(arg2)
				if not multiplayer
					chaoshudpos = arg2
				end
				player.chaos.chaoshudpos = arg2
			end
		end
		
		if not multiplayer 
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. chaoshud .."\x80 and \x87" .. chaoshudpos)
		else  
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. player.chaos.chaoshud .."\x80 and \x87" .. player.chaos.chaoshudpos)
		end
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		if not multiplayer 
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. chaoshud .."\x80 and \x87" .. chaoshudpos)
		else  
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. player.chaos.chaoshud .."\x80 and \x87" .. player.chaos.chaoshudpos)
		end
	end
end

local function ChaosCommand(player, arg1, arg2, arg3)
local cmdname = "chaoschances"
	if netgame and not IsPlayerAdmin(player)
		CONS_Printf(player, "\x85\ERROR:\x80 Can not use while in a netgame and not an admin!")
		return
	end
	
	local confirmchaoscmd
	
	if (tonumber(arg1) and (tonumber(arg1) >= 3 and tonumber(arg1) <= 9)) or arg1 == "default"
		if arg1 != "default"
			minwaittime = tonumber(arg1)
		else
			minwaittime = defaultminwaittime
		end
		
		confirmchaoscmd = true
	end
	
	if (tonumber(arg2) and (tonumber(arg2) >= 4 and tonumber(arg2) <= 20)) or arg2 == "default"
		if arg1 != "default"
			if tonumber(arg2) <= minwaittime
				CONS_Printf(player, "\x85\ERROR:\x80 Can't set maximum value below or equal to the current minimum value!")
			else
				maxwaittime = tonumber(arg2)
			end
		else
			maxwaittime = defaultmaxwaittime
		end
		
		confirmchaoscmd = true
	end
	
	if (arg3 == "normal" or arg3 == "always" or arg3 == "rarer") or arg3 == "default"
		if arg3 == "default" arg3 = "normal" end 
		if arg3 == "normal"
			vibechance = defaultvibechance
		elseif arg3 == "always"
			vibechance = FRACUNIT
		elseif arg3 == "rarer"
			vibechance = (FRACUNIT/2) - (FRACUNIT/9)
		end
		
		confirmchaoscmd = true
	end
	
	
	local vibechanceprint
	if vibechance == defaultvibechance
		vibechanceprint = "normal"
	elseif vibechance == FRACUNIT
		vibechanceprint = "always"
	else
		vibechanceprint = "rarer"
	end

	if confirmchaoscmd 
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
		CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. minwaittime ..", " .. maxwaittime .. ", " .. (vibechanceprint))
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. minwaittime ..", " .. maxwaittime .. ", " .. (vibechanceprint))
	end
end


local function ChaosPrint(player, arg1)
local cmdname = "chaosprint"
	if netgame and not IsPlayerAdmin(player)
		CONS_Printf(player, "\x85\ERROR:\x80 Can not use while in a netgame and not an admin!")
		return
	end
	
	if checkOnLean(arg1)
		chaosprint = checkOnLean(arg1, chaosprint)
		player.chaos.chaosprint = checkOnLean(arg1, player.chaos.chaosprint)
		changedvar(player, cmdname, arg1)
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		changedvar(player, cmdname, chaosprint) 
	end
end

local function ChaosBadMode(player, arg1)
local cmdname = "chaosbadmode"
	if netgame and not IsPlayerAdmin(player)
		CONS_Printf(player, "\x85\ERROR:\x80 Can not use while in a netgame and not an admin!")
		return
	end
	
	if checkOnLean(arg1)
		chaosbad = checkOnLean(arg1, chaosbad)
		changedvar(player, cmdname, arg1)
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		changedvar(player, cmdname, chaosbad) 
	end
end

local function ChaosEffectButton(player, arg1, arg2)
local cmdname = "chaoseffectbutton"
	local CMDCMPLTE
	if netgame -- somehow make this toggable via admin command 
		CONS_Printf(player, "\x85\ERROR:\x80 Can not use while in a netgame!")
		return
	end
	
	if KeyToNum(arg1) or arg1 == "off" then
		if not multiplayer
			CEbutton = arg1
		end
		player.chaos.CEbutton = arg1
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
		CMDCMPLTE = true
	end
	
	if (invtablesearch(statuseffects, arg2)) or arg2 == "none"
		if not multiplayer
			CEbuttonfx = arg2
		end
		player.chaos.CEbuttonfx = arg2
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
		CMDCMPLTE = true
	end
	
	if not CMDCMPLTE
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		if multiplayer 
			changedvar(player, cmdname, player.chaos.CEbutton) 
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. player.chaos.CEbutton .. "\x80 and \x87" .. player.chaos.CEbuttonfx)
		else 
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. CEbutton .. "\x80 and \x87" .. CEbuttonfx)
		end
		CONS_Printf(player, "*\x8E Options:\x80 off, weaponnext, weaponprev, attack/ringtoss, use/spin, camleft, camright, tossflag, firenormal/ringtossnormal, custom1, custom2, custom3, jump")
	else
		if multiplayer 
			changedvar(player, cmdname, player.chaos.CEbutton) 
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. player.chaos.CEbutton .. "\x80 and \x87" .. player.chaos.CEbuttonfx)
		else 
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. CEbutton .. "\x80 and \x87" .. CEbuttonfx)
		end
	end
end

local function ChaosBanEffect(player, arg1, arg2)
local cmdname = "chaosbaneffect"
	if netgame and not IsPlayerAdmin(player)
		CONS_Printf(player, "\x85\ERROR:\x80 Can not use while in a netgame and not an admin!")
		return
	end
	
	if arg1 and statuseffects[arg1] --and ((player.chaos.permeffect1 != arg1 and player.chaos.permeffect2 != arg1) and (permeffect1 != arg1 and permeffect2 != arg1) and not arg2 == "remove")
		if arg2 and arg2 == "remove"
			local removetime = tablesearch(bannedeffects, arg1)
			if removetime
				table.remove(bannedeffects, removetime) -- second value needs to be the number in the list where it's located in
				incrementeffects = $ - 1
			end
		else
			table.insert(bannedeffects, incrementeffects, arg1)
			incrementeffects = $ + 1
		end
		if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
	end
	CONS_Printf(player, "\x82" .. "Banned effects:")
	for key,value in pairs(bannedeffects) do CONS_Printf(player, "\x87* " .. value) end
	if incrementeffects == 1 CONS_Printf(player, "\x83" .. "None!") end
end

local function ChaosPermEffects(player, arg1, arg2)
local cmdname = "chaospermeffects"
	local checktime
	if (invtablesearch(effectduration, arg1) and tablesearch(effectduration, arg1) != 0 and arg1 != "change skin") or arg1 == "none"
		if arg1 == "magnet" and (someoneusingmagnet or multiplayer) return end
		if arg1 == "giant" and (player.chaos.permeffect1 == "tiny" or player.chaos.permeffect2 == "tiny") return end
		if arg1 == "tiny" and (player.chaos.permeffect1 == "giant" or player.chaos.permeffect2 == "giant") return end
		if arg1 != "none" and arg1 == player.chaos.permeffect2 return end
		if netgame and not IsPlayerAdmin(player)
			if tablesearch(bannedeffects, arg1) return end
			if tablesearch(bannedeffects, arg2) return end
		end
		
		if not multiplayer
			permeffect1 = arg1
		end
		
		if arg1 != player.chaos.permeffect1 and player.chaos.permeffect1 != "none"
			player.removepermeffect1 = player.chaos.permeffect1
		end
		
		player.chaos.permeffect1 = arg1
		player.checkpermeffect1 = true
		--if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
		checktime = true
	end
	
	if (invtablesearch(effectduration, arg2) and tablesearch(effectduration, arg2) != 0 and arg1 != "change skin") or arg2 == "none"
		if arg2 == "magnet" and (someoneusingmagnet or multiplayer) return end
		if arg2 == "giant" and (player.chaos.permeffect1 == "tiny" or player.chaos.permeffect2 == "tiny") return end
		if arg2 == "tiny" and (player.chaos.permeffect1 == "giant" or player.chaos.permeffect2 == "giant") return end
		if arg2 != "none" and arg2 == player.chaos.permeffect1 return end
		
		if not multiplayer
			permeffect2 = arg2
		end
		
		if arg2 != player.chaos.permeffect2 and player.chaos.permeffect2 != "none"
			player.removepermeffect2 = player.chaos.permeffect2
		end
		
		player.chaos.permeffect2 = arg2
		player.checkpermeffect2 = true
		--if not multiplayer and chaosautosave == "on" CH_SaveSettings(player) end
		checktime = true
	end
	
	if checktime
		if multiplayer
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. player.chaos.permeffect1 .."\x80 and \x87" .. player.chaos.permeffect2)
		else
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. permeffect1 .."\x80 and \x87" .. permeffect2)
		end
	else
		CONS_Printf(player, cmdDesc(commandDescription, cmdname))
		if multiplayer
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. player.chaos.permeffect1 .."\x80 and \x87" .. player.chaos.permeffect2)
		else
			CONS_Printf(player, "*\x8A " .. cmdname .. "\x80 is set to \x87" .. permeffect1 .."\x80 and \x87" .. permeffect2)
		end
	end
end

local function LoadSettingsAfter(player)
local cmdname = "chaosloadsettings"
	if multiplayer
		CONS_Printf(player, "\x85\ERROR:\x80 Can not use while in multiplayer!")
		return
	end
	CH_LoadSettings()
	CONS_Printf(player, "Successfully read Chaos Edition settings from file!")
end

local function SaveSettingsAfter(player)
local cmdname = "chaossavesettings"
	if multiplayer
		CONS_Printf(player, "\x85\ERROR:\x80 Can not use while in multiplayer!")
		return
	end
	CH_SaveSettings(player)
	CONS_Printf(player, "Successfully saved Chaos Edition settings to file!")
end

local function ChaosHelp(player, arg1) -- add if not multiplayer to hide some of the commands not relevant to multiplayer 
arg1 = tonumber(arg1)
local cmdname = "chaoshelp"

	if ReadChaosHelp == "no"
		ReadChaosHelp = "yes"
		CH_SaveSettings(player)
	end

	local maxpagenum = 2
	if arg1 == nil or arg1 <= 1 or arg1 > maxpagenum -- page 1
			CONS_Printf(player, "\x83\-- Page 1/" .. maxpagenum .. " : Main Commands --")
			--CONS_Printf(player, "*\x81 NOTICE:\x80 For commands that accept on/off, you can type \"toggle\" to switch between on/off.")
		CONS_Printf(player, cmdDesc(commandDescription, "chaoshelp"))
		if not netgame or (netgame and not IsPlayerAdmin(player)) CONS_Printf(player, cmdDesc(commandDescription, "chaosbadmode")) end
		if not netgame or (netgame and not IsPlayerAdmin(player)) CONS_Printf(player, cmdDesc(commandDescription, "chaoseffectbutton")) end
		if not netgame or (netgame and not IsPlayerAdmin(player)) CONS_Printf(player, cmdDesc(commandDescription, "chaosbaneffect")) end
		CONS_Printf(player, cmdDesc(commandDescription, "chaospermeffects"))
		if not netgame or (netgame and not IsPlayerAdmin(player)) CONS_Printf(player, cmdDesc(commandDescription, "chaoschances")) end
		CONS_Printf(player, cmdDesc(commandDescription, "chaoshud"))
		if not multiplayer CONS_Printf(player, cmdDesc(commandDescription, "chaosautoload")) end
		if not multiplayer CONS_Printf(player, cmdDesc(commandDescription, "chaosautosave")) end
	elseif arg1 == maxpagenum
			CONS_Printf(player, "\x83\-- Page 2/" .. maxpagenum .. " : Other Stuff --")
		CONS_Printf(player, cmdDesc(commandDescription, "chaosprint"))
		if not multiplayer CONS_Printf(player, cmdDesc(commandDescription, "chaossavesettings")) end
		if not multiplayer CONS_Printf(player, cmdDesc(commandDescription, "chaosloadsettings")) end
		if not netgame or (netgame and not IsPlayerAdmin(player)) CONS_Printf(player, cmdDesc(commandDescription, "chaosdebugger")) end
	end
end

-- [[ Add Commands ]] --
COM_AddCommand("chaosdebugger", ChaosDebugger, 1)
COM_AddCommand("chaoshud", ChaosHud, 0)
COM_AddCommand("chaosprint", ChaosPrint, 0)
COM_AddCommand("chaosautosave", ChaosAutosave, 0)
COM_AddCommand("chaosautoload", ChaosAutoload, 0)
COM_AddCommand("chaossavesettings", SaveSettingsAfter, 0)
COM_AddCommand("chaosloadsettings", LoadSettingsAfter, 0)
COM_AddCommand("chaoschances", ChaosCommand, 0)
COM_AddCommand("chaoshelp", ChaosHelp, 0)
COM_AddCommand("chaospermeffects", ChaosPermEffects, 0)
COM_AddCommand("chaosbadmode", ChaosBadMode, 0)
COM_AddCommand("chaosbaneffect", ChaosBanEffect, 0)
COM_AddCommand("chaoseffectbutton", ChaosEffectButton, 0)

-- trans rights --

-- [[[ The Actual Code ]]] --

addHook("PlayerThink", function (player)

	if not player.chaos.chaosloaded
		player.chaos.chaoshud = chaoshud
		player.chaos.chaosprint = chaosprint
		player.chaos.chaoshudpos = chaoshudpos
		player.chaos.chaosloaded = true
	end
	
	-- fix specifically for MMT menu
	if player.chaos.chaoshudpos and type(player.chaos.chaoshudpos) == "string"
		player.chaos.chaoshudpos = tonumber($)
	end

	if player.playerstate != PST_LIVE
		return
	end

	if not player.mo
		return
	end

	if player.bot > 0
		return
	end
	
	-- [[ MMT Save Stuff ]] --
	
	if not multiplayer and player.MMT
		if player.MMT.MMTModOptionChanged == "Chaos Edition"
			chaoshud = player.chaos.chaoshud
			CEbutton = player.chaos.CEbutton
			CEbuttonfx = player.chaos.CEbuttonfx
			chaoshudpos = player.chaos.chaoshudpos

			incrementeffects = tablelength(bannedeffects) + 1
			CH_SaveSettings(player)
		end
	end

	-- [[ the player functions ]] -- it's sloppy, but it works
	
	local function LavaCheck(sectorpic)
		if sectorpic == "GFZFLR02"
		or sectorpic == "GFZFLR21"
	
		or sectorpic == "CEZFLR02"
		or sectorpic == "CEZFLR03"
		or sectorpic == "STEEL01"
		or sectorpic == "STEEL2"
		or sectorpic == "THZFLR08"
	
		or sectorpic == "DS_ROCK1"
		or sectorpic == "DSFLOR1"
		or sectorpic == "DSFLOOR1"
		or sectorpic == "DS_TILE1"
	
		or sectorpic == "CEZGRS0"
		or sectorpic == "CEZGRS1"
		or sectorpic == "CECRPTR"
		or sectorpic == "CETILE2"
		or sectorpic == "CETILE2B"
		or sectorpic == "CESTONE5"
		or sectorpic == "CESTONE5"
		
		or sectorpic == "ACZFLR01"
		or sectorpic == "ACWOOD3F"
		
		or sectorpic == "ACWOOD3F"
		
		-- on MAP14 thus far
		
		
			return true
		else
			return false
		end
	
	end
	
			
	local function RemoveMeNow(effect)
		if effect == "giant"
			player.chaos.GiantRemoveNow = true
		elseif effect == "cement shoes"
			player.chaos.CementRemoveNow = true
		elseif effect == "tiny"
			player.chaos.TinyRemoveNow = true
		elseif effect == "no charability"
			player.chaos.CharAbilityRemoveNow = true
		elseif effect == "magnet"
			player.chaos.MagnetRemoveNow = true
		elseif effect == "drifting left"
			player.chaos.DriftingRemoveNow = true
		elseif effect == "butter physics"
			player.chaos.ButterRemoveNow = true
		elseif effect == "change skin color"
			player.chaos.SkinColorRemoveNow = true
		elseif effect == "change skin"
			player.chaos.SkinRemoveNow = true -- yowch
		elseif effect == "super jump"
			player.chaos.JumpRemoveNow = true
		elseif effect == "the ride never ends"
			player.chaos.SpinRemoveNow = true
		elseif effect == "the pizza is aggressive"
			player.chaos.PizzaRemoveNow = true
		elseif effect == "money bag"
			player.chaos.MoneyRemoveNow = true
		elseif effect == "earthquake"
			player.chaos.QuakeRemoveNow = true
		elseif effect == "gravity coil"
			player.chaos.GravityCoilRemoveNow = true
		elseif effect == "invert controls"
			player.chaos.InvertControlsRemoveNow = true
		elseif effect == "change sprite"
			player.chaos.ChangeSpriteRemoveNow = true
		elseif effect == "the floor is lava"
			player.chaos.LavaRemoveNow = true		
		elseif effect == "negative rings"
			player.chaos.NegativeRingsRemoveNow = true
		elseif effect == "darkened vision"
			player.chaos.DarkenedVisionRemoveNow = true
		elseif effect == "no turning"
			player.chaos.NoTurningRemoveNow = true
		elseif effect == "i was frozen today"
			player.chaos.FrozenRemoveNow = true
		elseif effect == "randomize sfx"
			player.chaos.RandomSFXRemoveNow = true		
		elseif effect == "black hole"
			player.chaos.BlackHoleRemoveNow = true

		end
	end
	
	local function IsPermanentEffect(arg1)
		if player.chaos.permeffect1 == arg1
			return true
		elseif player.chaos.permeffect2 == arg1
			return true
		else
			return false
		end
	end

	local function RemoveEffect(arg1, boo, perm)
		if boo != false EffectPrint(player, "*\x88\ expired\x80: " .. arg1) end
			if not perm
				if player.chaos.ContinuousEffect1 == arg1
					player.chaos.ContinuousEffect1 = nil
				elseif player.chaos.ContinuousEffect2 == arg1
					player.chaos.ContinuousEffect2 = nil
				elseif player.chaos.ContinuousEffect3 == arg1
					player.chaos.ContinuousEffect3 = nil
				elseif player.chaos.ContinuousEffect4 == arg1
					player.chaos.ContinuousEffect4 = nil
				end
			else
				if player.chaos.permeffect1 == arg1
					player.chaos.permeffect1 = nil
				elseif player.chaos.permeffect2 == arg1
					player.chaos.permeffect2 = nil
				end
			end
			
			if player.chaos.PreviousEffect1 == nil
				player.chaos.PreviousEffect1 = arg1
			elseif player.chaos.PreviousEffect2 == nil
				player.chaos.PreviousEffect2 = arg1
			elseif player.chaos.PreviousEffect3 == nil
				player.chaos.PreviousEffect3 = arg1
			elseif player.chaos.PreviousEffect4 == nil
				player.chaos.PreviousEffect4 = arg1
			else
			
			local CHOOSEONE = P_RandomRange(1, 4)
			if CHOOSEONE == 1
				player.chaos.PreviousEffect1 = arg1
				player.chaos.PE1Timer = 0
			elseif CHOOSEONE == 2
				player.chaos.PreviousEffect2 = arg1
				player.chaos.PE2Timer = 0
			elseif CHOOSEONE == 3
				player.chaos.PreviousEffect3 = arg1
				player.chaos.PE3Timer = 0
			else
				player.chaos.PreviousEffect4 = arg1
				player.chaos.PE4Timer = 0
			end
		end
	end
	
	-- [[ chaoseffectbutton ]] --
	-- maybe use LastCheck just to be sure it can be applied?
		if player.chaos.CEbutton and player.chaos.CEbutton != "off"
			if (player.cmd.buttons & KeyToNum(player.chaos.CEbutton)) and not player.chaos.HoldButton1 
				if player.chaos.CEbuttonfx == "none"
					player.chaos.vibecheck = true
				else
					if not CheckContiniousEffects(player, player.chaos.CEbuttonfx)
						ApplyEffect(player, player.chaos.CEbuttonfx)
					end
				end
				player.chaos.HoldButton1 = true
			end
				
			if player.cmd.buttons == player.cmd.buttons & ~KeyToNum(player.chaos.CEbutton)
				player.chaos.HoldButton1 = false
			end
		end

		
	-- [[ personal debugging tools ]] --
	
		--if player.cmd.buttons & BT_CUSTOM1 and chaosbeta == true
		--	print(mobjcount)
		--end
	
	-- [[ vibe check ]] --
	if not (player.pflags & PF_FINISHED) and not stoppedclock
		player.chaos.timer = $ + 1
		if not player.chaos.WaitTime
			player.chaos.WaitTime = P_RandomRange(minwaittime, maxwaittime)
		end
		if player.chaos.timer < TICRATE*player.chaos.WaitTime
			--debugprint("TIMER TIME")
		else
			debugconsprintf(player, "vibe check")
			player.chaos.timer = 0
			player.chaos.WaitTime = nil
			player.chaos.vibecheck = true
			player.chaos.RouletteTimer = P_RandomRange(1, TICRATE)
		end
	end
	
	if player.rerollTimer > 0
		player.rerollTimer = $ - 1
	end
	
	if player.chaos.HUDTEXT
		player.chaos.HUDTEXTTIMER = $ + 1
		if player.chaos.HUDTEXTTIMER == 3*TICRATE or player.chaos.HUDTEXT == "RESET"
			player.chaos.HUDTEXT = nil
			player.chaos.HUDTEXTTIMER = 0
			debugprint("hudtext timer expired")
		end
	end
	
	-- [[ remove permanent effects ]] --
	
	if player.removepermeffect1
		RemoveMeNow(player.removepermeffect1)
		debugprint("removed perm effect 1: " .. player.removepermeffect1)
		player.removepermeffect1 = nil
	end
	
	if player.removepermeffect2
		RemoveMeNow(player.removepermeffect2)
		debugprint("removed perm effect 2: " .. player.removepermeffect2)
		player.removepermeffect2 = nil
	end
	
	
	if player.checkpermeffect1
		local checker = CheckContiniousEffects(player, player.chaos.permeffect1)
		if checker and checker != "perm"
			RemoveMeNow(player.chaos.permeffect1)
		end
		
		if (player.chaos.permeffect1 == "giant" or player.chaos.permeffect2 == "giant") and CheckContiniousEffects(player, "tiny")
			RemoveMeNow("tiny")
		end
		
		if (player.chaos.permeffect1 == "tiny" or player.chaos.permeffect2 == "tiny") and CheckContiniousEffects(player, "giant")
			RemoveMeNow("giant")
		end
		
		player.checkpermeffect1 = false
	end	
	
	if player.checkpermeffect2
		local checker = CheckContiniousEffects(player, player.chaos.permeffect2)
		if checker and checker != "perm"
			RemoveMeNow(player.chaos.permeffect2)
		end
		
		if (player.chaos.permeffect1 == "giant" or player.chaos.permeffect2 == "giant") and CheckContiniousEffects(player, "tiny")
			RemoveMeNow("tiny")
		end
		
		if (player.chaos.permeffect1 == "tiny" or player.chaos.permeffect2 == "tiny") and CheckContiniousEffects(player, "giant")
			RemoveMeNow("giant")
		end
		
		player.checkpermeffect1 = false
	end
	
	-- [[ applying the status effects ]] --
	
	-- [ one-time fx ] --
	
	if CheckContiniousEffects(player, "anti-superring")
		S_StartSound(player.mo, sfx_antiri)
		--P_PlayerRingBurst(player, 10)
		
		if player.rings > 0
			if chaosbad == "on"
				P_GivePlayerRings(player, -20)
			else
				P_GivePlayerRings(player, -10)
			end
		end
		
		if player.rings == 0 and chaosbad != "on"
			player.powers[pw_flashing] = 1*TICRATE
		end
		
		RemoveEffect("anti-superring", false)
	end	
	
	if CheckContiniousEffects(player, "random music")
		--S_StartSound(player.mo, sfx_antiri)
		
		S_ChangeMusic(MusicRand(P_RandomRange(1, MusicRand("RANGE"))), true, player)
		
		RemoveEffect("random music", false)
	end	
	
	if CheckContiniousEffects(player, "spawn object")
		-- make sure object is not above ceiling, if it is, offset to spawn right below ceiling by using the object's height
		-- also have a random chance of spawning ahead or behind player 

		S_StartSound(player.mo, sfx_s3kcas)
		
		local RANDY1 = P_RandomRange(1, 15)
		local function Spawn(theobj, tic)
			local Obj = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z + (player.mo.height + (64*FRACUNIT)), theobj)
			local RANDY2 = P_RandomRange(1, 3)
			Obj.target = player.mo
			--Obj.tracer = player.mo
			Obj.angle = player.mo.angle
			if tic Obj.fuse = tic end
			if Obj.ceilingrover and Obj.ceilingrover.valid
				if Obj.z > Obj.ceilingrover.bottomheight
					Obj.z = Obj.ceilingrover.bottomheight - Obj.height - 4*FRACUNIT
				end
			else
				if Obj.z > Obj.subsector.sector.ceilingheight
					Obj.z = Obj.subsector.sector.ceilingheight - Obj.height - 4*FRACUNIT
				end
			end
			
			if RANDY2 == 1
				P_InstaThrust(Obj, Obj.angle, 22*FRACUNIT)
				Obj.angle = $ + ANGLE_180
			elseif RANDY2 == 3
				P_InstaThrust(Obj, Obj.angle, -22*FRACUNIT)
			end
		end
		
		-- include super duper rare chance of brak eggman spawning
		
		if RANDY1 == 1
			Spawn(MT_RING_BOX)
		elseif RANDY1 == 2
			Spawn(MT_BLUECRAWLA)
		elseif RANDY1 == 3
			Spawn(MT_GOLDBUZZ)
		elseif RANDY1 == 4
			Spawn(MT_JETTBOMBER)
		elseif RANDY1 == 5
			Spawn(MT_DETON)
		elseif RANDY1 == 6
			Spawn(MT_SPINCUSHION)
		elseif RANDY1 == 7
			Spawn(MT_FACESTABBER)
		elseif RANDY1 == 8
			Spawn(MT_BLUECRAWLA)
		elseif RANDY1 == 9
			Spawn(MT_EGGMOBILE)
		elseif RANDY1 == 10
			Spawn(MT_GOOMBA)
		elseif RANDY1 == 11
			Spawn(MT_REDSPRING)		
		elseif RANDY1 == 12
			Spawn(MT_TNTBARREL)		
		elseif RANDY1 == 13
			Spawn(MT_PYREFLY)
		elseif RANDY1 == 14
			Spawn(MT_PENGUINATOR)
		elseif RANDY1 == 15
			Spawn(MT_ROBOHOOD)
		end
		
		RemoveEffect("spawn object", false)
	end
	
	if CheckContiniousEffects(player, "superring")
		S_StartSound(player.mo, sfx_itemup)
		
		if chaosbad == "on"
			P_GivePlayerRings(player, 20)
		else
			P_GivePlayerRings(player, 10)
		end
		
		RemoveEffect("superring", false)
	end	
	
	if CheckContiniousEffects(player, "give good stuff")
		S_StartSound(player.mo, sfx_chchng)
		
		if not player.powers[pw_shield] and (P_RandomChance(FRACUNIT/2) or (chaosbad == "on"))
			player.powers[pw_shield] = SH_PITY
			P_SpawnShieldOrb(player)
			player.chaos.GivenNiceStuff = true
		end
		
		if P_RandomChance(FRACUNIT/2)
			player.powers[pw_sneakers] = 20*TICRATE
			--P_RestoreMusic(player)
			player.chaos.GivenNiceStuff = true
		end
		
		if P_RandomChance(FRACUNIT/3) or (chaosbad == "on" and P_RandomChance(FRACUNIT/2))
			player.powers[pw_invulnerability] = 20*TICRATE
			--P_RestoreMusic(player)
			player.chaos.GivenNiceStuff = true
		end
		
		if P_RandomChance(FRACUNIT/15) or (chaosbad == "on" and P_RandomChance(FRACUNIT/8))
			P_GivePlayerRings(player, 110)
			player.chaos.GivenNiceStuff = true
		end
		
		if P_RandomChance(FRACUNIT/99) or (chaosbad == "on" and P_RandomChance(FRACUNIT/70))
			P_GivePlayerRings(player, 999)
			player.powers[pw_invulnerability] = 500*TICRATE
			--P_RestoreMusic(player)
			player.chaos.GivenNiceStuff = true
		end
		
		if not player.chaos.GivenNiceStuff
			debugprint("somehow the player got nothing, giving them a bone")
			player.powers[pw_invulnerability] = 15*TICRATE
			P_GivePlayerRings(player, 15)
			--P_RestoreMusic(player)
		end
		
		player.chaos.GivenNiceStuff = nil
		
		RemoveEffect("give good stuff", false)
	end
	
	if CheckContiniousEffects(player, "i taketh")
		S_StartSound(player.mo, sfx_shldls)
		player.powers[pw_super] = 0
		P_RemoveShield(player)
		if player.rings > 0
			player.powers[pw_invulnerability] = 0
		end
		player.powers[pw_sneakers] = 0
		player.powers[pw_flashing] = 0
		player.powers[pw_tailsfly] = 0
		player.powers[pw_shield] = SH_NONE
		player.powers[pw_nights_superloop] = 0
		player.powers[pw_nights_helper] = 0
		player.powers[pw_nights_linkfreeze] = 0
		if player.rings > 0
			if chaosbad == "on"
				P_GivePlayerRings(player, -7)
			else
				P_GivePlayerRings(player, -3)
			end
		end
		player.glidetime = 0
		
		if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS)
			if player.drillmeter > 0
				player.drillmeter = $ / 2
			elseif player.chaos.OGdrillmeter > 0
				player.chaos.OGdrillmeter = $ / 2
			end
			if player.spheres > 2
				player.spheres = $ - 2
			end
		end

		player.climbing = 0
		player.homing = 0
		player.mo.color = player.skincolor
		player.pflags = $1 | PF_THOKKED
		player.pflags = $ & ~PF_GLIDING
		player.pflags = $ & ~PF_SPINNING
		--P_RestoreMusic(player)
		RemoveEffect("i taketh", false)
	end	
	
	if CheckContiniousEffects(player, "nuke")
		if player.powers[pw_shield] 
			player.chaos.NukePrevShield = player.powers[pw_shield] 
		end
		
		P_BlackOw(player)
		
		if player.chaos.NukePrevShield 
			player.powers[pw_shield] = player.chaos.NukePrevShield
			player.chaos.NukePrevShield = nil
		end
		RemoveEffect("nuke", false)
	end
	
	
	if CheckContiniousEffects(player, "fling player")
		-- chcek the vibe check place for the random chance of slapping, couldn't do it any other way 
		local RANDY = P_RandomChance(FRACUNIT/2)
		
		if chaosbad == "on"
			if RANDY
				P_InstaThrust(player.mo, player.mo.angle, 310*FRACUNIT)
			else
				P_InstaThrust(player.mo, player.mo.angle, -290*FRACUNIT)
			end
		else
			if RANDY
				P_InstaThrust(player.mo, player.mo.angle, 210*FRACUNIT)
			else
				P_InstaThrust(player.mo, player.mo.angle, -190*FRACUNIT)
			end
		end
		
		if player.chaos.Slapped
			player.mo.state = S_PLAY_PAIN
			player.climbing = 0
			if chaosbad == "on"
				player.mo.momz = $ + 35*FRACUNIT
			else
				player.mo.momz = $ + 20*FRACUNIT
			end
			S_StartSound(player.mo, sfx_slappy)
			--player.chaos.HUDTEXT2 = "slapped"
			--player.chaos.HUDTEXTChange = true
		else
			S_StartSound(player.mo, sfx_s3k81)
		end
		
		RemoveEffect("fling player", false)
	end	
	
	if player.chaos.Slapped and chaosbad != "on"
		player.powers[pw_flashing] = TICRATE
		if P_IsObjectOnGround(player.mo)
			player.chaos.Slapped = false
		end
	end
	
	if CheckContiniousEffects(player, "remove random effect") -- repeat until
		
		repeat
		local ChanceForRemoval = P_RandomRange(1, 4)
		if ChanceForRemoval == 1 -- and not itself
			if player.chaos.ContinuousEffect1 and not (player.chaos.ContinuousEffect1 == "remove random effect")
				RemoveMeNow(player.chaos.ContinuousEffect1)
				player.chaos.SuccessfulRemoval = true
			end
		elseif ChanceForRemoval == 2 
			if player.chaos.ContinuousEffect2 and not (player.chaos.ContinuousEffect2 == "remove random effect")
				RemoveMeNow(player.chaos.ContinuousEffect2)
				player.chaos.SuccessfulRemoval = true
			end
		elseif ChanceForRemoval == 3 
			if player.chaos.ContinuousEffect3 and not (player.chaos.ContinuousEffect3 == "remove random effect")
				RemoveMeNow(player.chaos.ContinuousEffect3)
				player.chaos.SuccessfulRemoval = true
			end
		elseif ChanceForRemoval == 4
			if player.chaos.ContinuousEffect4 and not (player.chaos.ContinuousEffect4 == "remove random effect")
				RemoveMeNow(player.chaos.ContinuousEffect4)
				player.chaos.SuccessfulRemoval = true
			end
		end
		ChanceForRemoval = $ + 1
		if ChanceForRemoval > 4
			ChanceForRemoval = 1
		end
		until player.chaos.SuccessfulRemoval == true
		
		if player.chaos.SuccessfulRemoval
			RemoveEffect("remove random effect", false)
			player.chaos.SuccessfulRemoval = false
		end
	end
	
	-- [ timer fx, starting section ] --
	
	if not player.chaos.DriftingNow
		if CheckContiniousEffects(player, "drifting left")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.DriftingNow = true
			if chaosbad == "off"
				player.chaos.DriftingTimer = 3*TICRATE
			end
		end
	else
		local effecttime = invtablesearch(effectduration, "drifting left")*TICRATE
		
		if not IsPermanentEffect("drifting left")
			player.chaos.DriftingTimer = $ + 1
		end

		if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS)
			player.flyangle = $ + 2
			player.anotherflyangle = $ + 2
		else
			if chaosbad == "on"
				player.mo.angle = (player.cmd.angleturn << FRACBITS) + ANG2 + ANG2
			else
				player.mo.angle = (player.cmd.angleturn << FRACBITS) + ANG2
			end
			--player.cmd.angleturn = $<<16 + ANG2
		end
		
		if player.chaos.DriftingTimer >= effecttime or player.chaos.DriftingRemoveNow
			RemoveEffect("drifting left")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.DriftingTimer = 0
			player.chaos.DriftingRemoveNow = false
			player.chaos.DriftingNow = false
		end
	end
	
	if not player.chaos.QuakeNow
		if CheckContiniousEffects(player, "earthquake")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.QuakeNow = true
			if netgame chatprint("* " .. player.name .. " has started an earthquake!", true) end
		end
	else
		local effecttime = invtablesearch(effectduration, "earthquake")*TICRATE
		
		if not IsPermanentEffect("earthquake")
			player.chaos.QuakeTimer = $ + 1
		end

		if chaosbad == "on"
			P_StartQuake(30*FRACUNIT, 1, nil, nil)
		else
			P_StartQuake(20*FRACUNIT, 1, nil, nil)
		end
		
		if player.chaos.QuakeTimer >= effecttime or player.chaos.QuakeRemoveNow
			RemoveEffect("earthquake")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.QuakeTimer = 0
			player.chaos.QuakeRemoveNow = false
			player.chaos.QuakeNow = false
		end
	end
	
	if not player.chaos.MoneyBagNow
		if CheckContiniousEffects(player, "money bag")
			S_StartSound(player.mo, sfx_s3kb3)
			player.chaos.MoneyBagNow = true
			local RANDYC = P_RandomRange(0, TICRATE+(TICRATE)/2) -- approximately 17 tics, or half a second
			player.chaos.MoneyTimer = $ - RANDYC
			debugprint(RANDYC)
		end
	else
		local effecttime = invtablesearch(effectduration, "money bag")*TICRATE
	
		if not IsPermanentEffect("money bag")
			player.chaos.MoneyTimer = $ + 1
		end
		
		if chaosbad == "on"
			if (player.chaos.MoneyTimer % 5 == 0)
				if G_IsSpecialStage(gamemap)
					player.spheres = $ + P_RandomRange(1, 2)
					S_StartSound(player.mo, sfx_s3k65)
				else
					P_GivePlayerRings(player, P_RandomRange(1, 2))
					S_StartSound(player.mo, sfx_itemup)
				end
			end		
		else
			if (player.chaos.MoneyTimer % 5 == 0)
				if G_IsSpecialStage(gamemap)
					player.spheres = $ + 1
					S_StartSound(player.mo, sfx_s3k65)
				else
					P_GivePlayerRings(player, 1)
					S_StartSound(player.mo, sfx_itemup)
				end
			end
		end
		
		if player.chaos.MoneyTimer >= effecttime or player.chaos.MoneyRemoveNow
			RemoveEffect("money bag")
			S_StartSound(player.mo, sfx_chchng)
			player.chaos.MoneyTimer = 0
			player.chaos.MoneyRemoveNow = false
			player.chaos.MoneyBagNow = false
		end
	end
	
	if not player.chaos.GhostNow
		if CheckContiniousEffects(player, "ghost")
			S_StartSound(player.mo, sfx_boo)
			player.chaos.GhostNow = true
		end
	else
		local effecttime = invtablesearch(effectduration, "ghost")*TICRATE
		
		if not IsPermanentEffect("ghost")
			player.chaos.GhostTimer = $ + 1
		end

		player.mo.flags = $ | MF_NOCLIPTHING
		--player.mo.flags2 = $ | MF2_SHADOW
		player.pflags = $1 | PF_INVIS
		
		if player.chaos.GhostTimer >= effecttime or player.chaos.GhostRemoveNow
			RemoveEffect("ghost")
			S_StartSound(player.mo, sfx_kc65)
			player.mo.flags = $ & ~MF_NOCLIPTHING
			--player.mo.flags2 = $ & ~MF2_SHADOW
			player.pflags = $ & ~PF_INVIS
			player.chaos.GhostTimer = 0
			player.chaos.GhostRemoveNow = false
			player.chaos.GhostNow = false
			player.chaos.GhostBegone = true
			player.powers[pw_flashing] = 1*TICRATE
		end
	end
	
	if not player.chaos.PizzaNow -- should've been pizzatime
		if CheckContiniousEffects(player, "the pizza is aggressive")
			S_StartSound(player.mo, sfx_chapis)
			player.chaos.PizzaNow = true
			if chaosbad == "off"
				player.chaos.PizzaTimer = 3*TICRATE
			end
		end
	else
		local effecttime = invtablesearch(effectduration, "the pizza is aggressive")*TICRATE
		
		if not player.chaos.PizzaSpawned
			player.chaos.PizzaRemoval = nil
			local thepizzais = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.ceilingz-(6*FRACUNIT), MT_JIMBO)
			thepizzais.target = player.mo
			
			local size
			local choosesize = P_RandomRange(1, 4)
			
			if chaosbad == "on"
				choosesize = P_RandomRange(2, 5)
			end
			
			if choosesize == 1
				size = "small"
			elseif choosesize == 2 
				size = "medium"
			elseif choosesize == 3
				if P_RandomChance(FRACUNIT/2)
					size = "large"
				else
					if P_RandomChance(FRACUNIT/2)
						size = "medium"
					else
						size = "small"
					end
				end
			elseif choosesize == 4
				if P_RandomChance(FRACUNIT/99) -- extrodionarly rare chance!
					size = "extra large"
				else
					if P_RandomChance(FRACUNIT/2 + FRACUNIT/3)
						size = "medium"
					else
						if P_RandomChance(FRACUNIT/2)
							size = "small"
						else
							size = "large"
						end
					end
				end
			elseif choosesize == 5 -- only happens in bad mode
				size = "extra large"
			end
			
			if not size -- failsafe, just for in case of some really weird situation occuring
				size = "small"
			end
			
			if size == "large" or size == "medium" or size == "small" or size == "personal" or size == "xl" or size == "extra large"  or size == "extralarge" -- REMOVE THIS LATER
				thepizzais.sizeoption = size
			end
			
			local toppings = "pepperoni"
			
			if P_RandomChance(FRACUNIT/4)
				thepizzais.state = S_CHAOSEXPLOSPIZZA
				toppings = "deton"
			end
			
			print(player.name .. " has ordered a " .. size .. " sized pizza, with pepperoni toppings to themselves.")
			if netgame chatprint(player.name .. " has ordered a " .. size .. " sized pizza, with " .. toppings .. " toppings to themselves.", true) end
			player.chaos.PizzaSpawned = true
		end
		
		if not IsPermanentEffect("the pizza is aggressive")
			player.chaos.PizzaTimer = $ + 1
		else
			player.chaos.PizzaTimer = 2
		end
		
		if player.chaos.PizzaTimer >= effecttime or player.chaos.PizzaRemoveNow
			RemoveEffect("the pizza is aggressive")
			if not player.chaos.PizzaHitTarget 
				S_StartSound(player.mo, sfx_kc65)
			end
			player.chaos.PizzaTimer = 0
			player.chaos.PizzaRemoval = true
			player.chaos.PizzaRemoveNow = false
			player.chaos.PizzaHitTarget = false
			player.chaos.PizzaNow = false
			player.chaos.PizzaSpawned = false
		end
	end
	
	if not player.chaos.SkinColorNow
		if CheckContiniousEffects(player, "change skin color")
			player.chaos.SkinColorNow = true
		end
	else
		local effecttime = invtablesearch(effectduration, "change skin color")*TICRATE
	
		if not IsPermanentEffect("change skin color")
			player.chaos.SkinColorTimer = $ + 1
		end
		
		if (player.chaos.SkinColorTimer % 4 == 0) 
			player.mo.color = P_RandomRange(1, #skincolors-1)
			S_StartSound(player.mo, sfx_kc39)
			player.skincolor = player.mo.color
		end
		
		if player.chaos.SkinColorTimer > effecttime or player.chaos.SkinColorRemoveNow
			RemoveEffect("change skin color")
			S_StartSound(player.mo, sfx_kc65)
			local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST) -- stole the x and y coordinates stuff from the SRB2 source code; specifically what's used for tails' followitem
			dust.scale = FRACUNIT*2
			dust.state = S_XPLD1
			player.chaos.SkinColorTimer = 0
			player.chaos.SkinColorRemoveNow = false
			player.chaos.SkinColorNow = false
		end
	end
	
	if not player.chaos.SpinNow
		if CheckContiniousEffects(player, "the ride never ends")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.SpinNow = true
			local RANDYSPIN = P_RandomChance(FRACUNIT/2)
			if RANDYSPIN
				debugprint("spin regular")			
			else
				debugprint("spin other")
				player.chaos.SpinOtherWay = true
			end
			if chaosbad == "off"
				player.chaos.SpinTimer = 3*TICRATE
			end
		end
	else
		local effecttime = invtablesearch(effectduration, "the ride never ends")*TICRATE
		
		if not IsPermanentEffect("the ride never ends")
			player.chaos.SpinTimer = $ + 1
		end
		
		if chaosbad == "on"
			if not player.chaos.SpinOtherWay
				if player.viewrollangle < (ANG30 + ANG30)
					player.viewrollangle = $ + (ANG1/4)
					player.mo.rollangle = $ - (ANG1/4)
				end
			else
				if player.viewrollangle > -(ANG30 + ANG30)
					player.viewrollangle = $ - (ANG1/4)
					player.mo.rollangle = $ + (ANG1/4)
				end
			end
		else		
			if not player.chaos.SpinOtherWay
				if player.viewrollangle < (ANG30 + ANG15)
					player.viewrollangle = $ + (ANG1/5)
					player.mo.rollangle = $ - (ANG1/5)
				end
			else
				if player.viewrollangle > -(ANG30 + ANG15)
					player.viewrollangle = $ - (ANG1/5)
					player.mo.rollangle = $ + (ANG1/5)
				end
			end
		end
		
		if player.chaos.SpinTimer >= effecttime or player.chaos.SpinRemoveNow
			RemoveEffect("the ride never ends")
			S_StartSound(player.mo, sfx_kc65)
			player.chaos.RestoreViewAngle = true
			player.chaos.SpinRemoveNow = false
			player.chaos.SpinNow = false
			player.chaos.SpinTimer = 0
			player.chaos.RANDYSPIN = nil
		end
	end
	
	if not player.chaos.SkinNow
		if CheckContiniousEffects(player, "change skin")
			player.chaos.SkinNow = true
			local RANDYC = P_RandomRange(-5, TICRATE-(TICRATE)/2) -- approximately 17 tics, or half a second  
			player.chaos.SkinTimer = $ + RANDYC
			debugprint(RANDYC)
		end
	else
		local effecttime = invtablesearch(effectduration, "change skin")*TICRATE
		local skinnum = P_RandomRange(0, numberofskins())
		
		player.chaos.SkinTimer = $ + 1
		
		if (player.chaos.SkinTimer % 4 == 0) -- P_RandomRange(4, 6)
			if not R_SkinUsable(player, skinnum)
				skinnum = P_RandomRange(0, 2)
			end
			S_StartSound(player.mo, sfx_kc39)
			R_SetPlayerSkin(player, skinnum)
		end
		
		if player.chaos.SkinTimer > effecttime or player.chaos.SkinRemoveNow
			RemoveEffect("change skin")
			local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST) 
			dust.scale = FRACUNIT*2
			dust.state = S_XPLD1
			S_StartSound(player.mo, sfx_kc65)
			player.chaos.SkinTimer = 0
			player.chaos.SkinRemoveNow = false
			player.chaos.SkinNow = false
		end
	end
	
	if not player.chaos.ButterNow
		if CheckContiniousEffects(player, "butter physics")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.ButterNow = true
			if chaosbad == "off"
				player.chaos.ButterTimer = 3*TICRATE
			end
		end
	else
		local effecttime = invtablesearch(effectduration, "butter physics")*TICRATE
	
		if not IsPermanentEffect("butter physics")
			player.chaos.ButterTimer = $ + 1
		end
		
		if chaosbad == "on"
			player.mo.friction = FRACUNIT + (FRACUNIT/69) -- nice
		else
			player.mo.friction = FRACUNIT + (FRACUNIT/55)
		end
		
		if player.chaos.ButterTimer >= effecttime or player.chaos.ButterRemoveNow
			RemoveEffect("butter physics")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.ButterTimer = 0
			player.mo.friction = player.chaos.OGFriction
			player.chaos.ButterRemoveNow = false
			player.chaos.ButterNow = false
		end
	end
	
	if not player.chaos.GiantNow
		if CheckContiniousEffects(player, "giant") and (player.powers[pw_carry] != CR_MINECART and player.powers[pw_carry] != CR_ROLLOUT)
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.OGGiantScale = player.mo.scale
			player.mo.destscale = (FRACUNIT*4)
			player.chaos.GiantNow = true
		end
	else
		local effecttime = invtablesearch(effectduration, "giant")*TICRATE
		
		if (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_ROLLOUT)
			debugprint("giant but in minecart or rollout rock, aborting")
			player.chaos.GiantRemoveNow = true
		end
	
		if not IsPermanentEffect("giant")
			player.chaos.GiantTimer = $ + 1
		end
		
		if player.chaos.GiantTimer >= effecttime or player.chaos.GiantRemoveNow
			RemoveEffect("giant")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.GiantTimer = 0
			player.mo.destscale = FRACUNIT
			player.chaos.GiantRemoveNow = false
			player.chaos.GiantNow = false
		end
	end		
	
	if not player.chaos.RandomSFXNow
		if CheckContiniousEffects(player, "randomize sfx")
			S_StartSound(player.mo, sfx_kc5f)
			player.chaos.RandomSFXNow = true
		end
	else
		local effecttime = invtablesearch(effectduration, "randomize sfx")*TICRATE
	
		if not IsPermanentEffect("randomize sfx")
			player.chaos.RandomSFXTimer = $ + 1
		end
		
		if player.chaos.RandomSFXTimer >= effecttime or player.chaos.GiantRemoveNow
			RemoveEffect("randomize sfx")
			S_StartSound(player.mo, sfx_kc65)
			player.chaos.RandomSFXTimer = 0
			player.chaos.RandomSFXRemoveNow = false
			player.chaos.RandomSFXNow = false
		end
	end	
	
	if not player.chaos.TinyNow
		if CheckContiniousEffects(player, "tiny") and (player.powers[pw_carry] != CR_MINECART and player.powers[pw_carry] != CR_ROLLOUT)
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.OGTinyScale = player.mo.scale
			if chaosbad == "on"
				player.mo.destscale = (FRACUNIT/8)
			else
				player.mo.destscale = (FRACUNIT/6)
			end
			player.chaos.TinyNow = true
		end
	else
		local effecttime = invtablesearch(effectduration, "tiny")*TICRATE
		
		if (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_ROLLOUT)
			debugprint("tiny but in minecart or rollout rock, aborting")
			player.chaos.TinyRemoveNow = true
		end

		if not IsPermanentEffect("tiny")
			player.chaos.TinyTimer = $ + 1
		end
		
		if player.chaos.TinyTimer >= effecttime or player.chaos.TinyRemoveNow
			RemoveEffect("tiny")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.TinyTimer = 0
			player.mo.destscale = FRACUNIT
			player.chaos.TinyRemoveNow = false
			player.chaos.TinyNow = false
		end
	end	
	
	if not player.chaos.FunkyCameraNow
		if CheckContiniousEffects(player, "funky camera")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.FunkyCameraNow = true
			
			local RANDY = P_RandomChance(FRACUNIT/2)
			
			if RANDY
				player.chaos.CameraDist = true
			else
				player.chaos.CameraDist = false
			end
			
		end
	else
		local effecttime = invtablesearch(effectduration, "funky camera")*TICRATE
		
		if chaosbad == "on"
			if player.chaos.CameraDist
				player.camerascale = skins[player.mo.skin].camerascale * 7
			else
				player.camerascale = skins[player.mo.skin].camerascale / 9
			end
		else
			if player.chaos.CameraDist
				player.camerascale = skins[player.mo.skin].camerascale * 4
			else
				player.camerascale = skins[player.mo.skin].camerascale / 5
			end
		end
		
		if not IsPermanentEffect("funky camera")
			player.chaos.FunkyCameraTimer = $ + 1
		end
		
		if player.chaos.FunkyCameraTimer >= effecttime or player.chaos.FunkyCameraRemoveNow
			RemoveEffect("funky camera")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.FunkyCameraTimer = 0
			player.camerascale = skins[player.mo.skin].camerascale
			player.chaos.CameraDist = nil
			player.chaos.FunkyCameraRemoveNow = false
			player.chaos.FunkyCameraNow = false
		end
	end	
	
	if not player.chaos.LavaNow
		if CheckContiniousEffects(player, "the floor is lava")
			S_StartSound(player.mo, sfx_s1ae) -- hopefully that fits
			player.chaos.LavaNow = true	
			if chaosbad == "off"
				player.chaos.LavaTimer = 5*TICRATE
			end
		end
	else
		local effecttime = invtablesearch(effectduration, "the floor is lava")*TICRATE
		
		--if chaosbad == "on"
			if P_IsObjectOnGround(player.mo) and not (player.powers[pw_shield] & SH_PROTECTFIRE)
				P_DamageMobj(player.mo)
			end
		--[[else
			if P_IsObjectOnGround(player.mo) --to-do: reverse gravity check, 
			-- check to make sure player is at the same height w/ the sector with the texture 
				if player.mo.floorrover
					if LavaCheck(player.mo.floorrover.sector.ceilingpic)
						P_DamageMobj(player.mo)
					end
				else
					if LavaCheck(player.mo.subsector.sector.floorpic)
						P_DamageMobj(player.mo)
					end
				end
			end
		end]]
		
		if not IsPermanentEffect("the floor is lava")
			player.chaos.LavaTimer = $ + 1
		end
		
		if player.chaos.LavaTimer >= effecttime or player.chaos.FunkyCameraRemoveNow
			RemoveEffect("the floor is lava")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.LavaTimer = 0
			player.chaos.LavaRemoveNow = false
			player.chaos.LavaNow = false
		end
	end	
	
	if not player.chaos.NegativeRingsNow
		if CheckContiniousEffects(player, "negative rings")
			S_StartSound(player.mo, sfx_kc59) -- change to anti-ring sfx
			player.chaos.NegativeRingsNow = true
			if chaosbad == "off"
				player.chaos.NegativeRingsTimer = 4*TICRATE
			end
		end
	else
		local effecttime = invtablesearch(effectduration, "negative rings")*TICRATE
		
		if not IsPermanentEffect("negative rings")
			player.chaos.NegativeRingsTimer = $ + 1
		end
		
		if player.chaos.NegativeRingsTimer >= effecttime or player.chaos.NegativeRingsRemoveNow
			RemoveEffect("negative rings")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.NegativeRingsTimer = 0
			player.chaos.NegativeRingsRemoveNow = false
			player.chaos.NegativeRingsNow = false
		end
	end	
	
	if not player.chaos.FrozenNow -- if implementing the ice block from sonic mania, use sfx_iceb 
		if CheckContiniousEffects(player, "i was frozen today")
			S_StartSound(player.mo, sfx_s3k80) -- change to sfx for ice in sonic mania 
			player.chaos.FrozenNow = true
			player.chaos.FrozenPrevColor = player.mo.color
		end
	else
		local effecttime = invtablesearch(effectduration, "i was frozen today")*TICRATE

		--player.mo.state = S_PLAY_PAIN
		player.mo.colorized = true
		player.mo.color = SKINCOLOR_ICY
		if not player.chaos.FrozenState
			player.chaos.FrozenState = player.mo.state
			player.chaos.FrozenFrame = player.mo.frame
			if player.followmobj
				player.chaos.FollowFrozenState = player.followmobj.state
				player.chaos.FollowFrozenFrame = player.followmobj.frame
			end
		end
		
		
		if not IsPermanentEffect("i was frozen today")
			player.chaos.FrozenTimer = $ + 1
		end
		
		if player.chaos.FrozenTimer >= effecttime or player.chaos.FrozenRemoveNow
			RemoveEffect("i was frozen today")
			S_StartSound(player.mo, sfx_s3k80)
			player.mo.color = player.chaos.FrozenPrevColor
			player.mo.colorized = false
			player.chaos.FrozenState = nil
			player.chaos.FollowFrozenState = nil			
			player.chaos.FrozenFrame = nil
			player.chaos.FollowFrozenFrame = nil
			player.chaos.FrozenTimer = 0
			player.chaos.FrozenRemoveNow = false
			player.chaos.FrozenNow = false
		end
	end
	
	if not player.chaos.DarkenedVisionNow
		if CheckContiniousEffects(player, "darkened vision")
			S_StartSound(player.mo, sfx_kc59) -- change to anti-ring sfx
			player.chaos.DarkenedVisionNow = true
		end
	else
		local effecttime = invtablesearch(effectduration, "darkened vision")*TICRATE
		
		if not IsPermanentEffect("darkened vision")
			player.chaos.DarkenedVisionTimer = $ + 1
		end
		
		if player.chaos.DarkenedVisionTimer >= effecttime or player.chaos.DarkenedVisionRemoveNow
			RemoveEffect("darkened vision")
			S_StartSound(player.mo, sfx_kc65)
			player.chaos.DarkenedVisionTimer = 0
			player.chaos.DarkenedVisionRemoveNow = false
			player.chaos.DarkenedVisionNow = false
		end
	end	
	
	
	if not player.chaos.NoTurningNow
		if CheckContiniousEffects(player, "no turning")
			S_StartSound(player.mo, sfx_kc59) -- change to downscale sfx if not already that 
			player.chaos.NoTurningNow = true
		end
	else
		local effecttime = invtablesearch(effectduration, "no turning")*TICRATE
		
		if not player.chaos.NoTurningAngle
			player.chaos.NoTurningAngle = player.mo.angle
		end
		
		player.mo.angle = player.chaos.NoTurningAngle
		
		if not IsPermanentEffect("no turning")
			player.chaos.NoTurningTimer = $ + 1
		end
		
		if player.chaos.NoTurningTimer >= effecttime or player.chaos.NoTurningRemoveNow
			RemoveEffect("no turning")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.NoTurningTimer = 0
			player.chaos.NoTurningRemoveNow = false
			player.chaos.NoTurningNow = false
			player.chaos.NoTurningAngle = nil
		end
	end

	if not player.chaos.CementNow
		if CheckContiniousEffects(player, "cement shoes")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.CementNow = true
			player.chaos.CementPrevColor = player.mo.color
			if player.powers[pw_shield] & SH_PROTECTSPIKE
				player.chaos.ShieldAlreadyHasSpike = true
			end
		end
	else -- sfx_s3k51
		local effecttime = invtablesearch(effectduration, "cement shoes")*TICRATE
		
		player.mo.colorized = true
		
		player.mo.color = SKINCOLOR_GREY
		
		if not player.chaos.ShieldAlreadyHasSpike and chaosbad != "on"
			player.powers[pw_shield] = $ | SH_PROTECTSPIKE
		end
		
		if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS)
			player.mo.momx = $ / 3
			player.mo.momy = $ / 3
			player.mo.momz = $ / 4
		else
			if chaosbad == "on"
				player.normalspeed = skins[player.mo.skin].normalspeed / 7
				player.runspeed = skins[player.mo.skin].runspeed / 5
				player.actionspd = skins[player.mo.skin].actionspd / 9
				player.thrustfactor = skins[player.mo.skin].thrustfactor / 5
				player.mindash = FRACUNIT
				player.maxdash = skins[player.mo.skin].maxdash / 3
			else
				player.normalspeed = skins[player.mo.skin].normalspeed / 5
				player.runspeed = skins[player.mo.skin].runspeed / 4
				player.actionspd = skins[player.mo.skin].actionspd / 7
				player.thrustfactor = skins[player.mo.skin].thrustfactor / 3
			end
			
			-- P_SetObjectMomZ(player.mo, -gravity, true)
			
			if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP)
				if player.mo.momz > 0
					player.mo.momz = $ + $ / 10
				else
					player.mo.momz = $ - $ / 8
				end
			else
				if player.mo.momz > 0
					player.mo.momz = $ - $ / 10
				else
					player.mo.momz = $ + $ / 8
				end
			end
		end
	
		if not IsPermanentEffect("cement shoes")
			player.chaos.CementTimer = $ + 1
		end
		
		if not P_IsObjectOnGround(player.mo)
			player.chaos.CementShoesSfx = true
			if player.mo.momz < -(3*FRACUNIT) and not player.chaos.CementShoesFallSfx
				S_StartSound(player.mo, sfx_s3k51)
				player.chaos.CementShoesFallSfx = true
			end
		end
		
		if P_IsObjectOnGround(player.mo)
			if player.chaos.CementShoesSfx
				S_StartSound(player.mo, sfx_pstop) --previously sfx_doord2
				if not netgame or (netgame and numberinserver() <= 2)
					P_StartQuake(10*FRACUNIT, 5, nil, nil)
				end
				player.chaos.CementShoesSfx = false
			end
			player.chaos.CementShoesFallSfx = false
		end
		
		if player.chaos.CementTimer >= effecttime or player.chaos.CementRemoveNow
			RemoveEffect("cement shoes")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.CementTimer = 0
			player.normalspeed = skins[player.mo.skin].normalspeed
			player.runspeed = skins[player.mo.skin].runspeed
			player.actionspd = skins[player.mo.skin].actionspd
			player.thrustfactor = skins[player.mo.skin].thrustfactor
			player.mindash = skins[player.mo.skin].mindash
			player.maxdash = skins[player.mo.skin].maxdash
			player.mo.color = player.chaos.CementPrevColor
			player.mo.colorized = false
			player.chaos.CementRemoveNow = false
			player.chaos.CementNow = false
			player.chaos.CementShoesSfx = false
			player.chaos.CementShoesFallSfx = false
			if not player.chaos.ShieldAlreadyHasSpike
				player.powers[pw_shield] = $ & ~SH_PROTECTSPIKE
			end
			P_SpawnShieldOrb(player)
			player.chaos.ShieldAlreadyHasSpike = nil
		end
	end
	
	if not player.chaos.SuperJumpNow
		if CheckContiniousEffects(player, "super jump")
			player.chaos.SuperJumpNow = true
			S_StartSound(player.mo, sfx_kc5a)
		end
	else
		local effecttime = invtablesearch(effectduration, "super jump")*TICRATE
		
		if chaosbad == "on"
			player.jumpfactor = skins[player.mo.skin].jumpfactor * 5
		else
			player.jumpfactor = skins[player.mo.skin].jumpfactor * 4
		end
		
		if not IsPermanentEffect("super jump")
			player.chaos.JumpTimer = $ + 1
		end
		
		if player.chaos.JumpTimer >= effecttime or player.chaos.JumpRemoveNow
			RemoveEffect("super jump")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.JumpTimer = 0
			player.jumpfactor = skins[player.mo.skin].jumpfactor
			player.chaos.JumpRemoveNow = false
			player.chaos.SuperJumpNow = false
		end
	
	end
	
	if not player.chaos.GravityCoilNow
		if CheckContiniousEffects(player, "gravity coil")
			player.chaos.GravityCoilNow = true
			S_StartSound(player.mo, sfx_chgc)
		end
	else
		local effecttime = invtablesearch(effectduration, "gravity coil")*TICRATE
		
		if skins[player.mo.skin].jumpfactor >= FRACUNIT
			player.jumpfactor = (skins[player.mo.skin].jumpfactor + (FRACUNIT/2) + (FRACUNIT/4))
		else
			player.jumpfactor = (FRACUNIT + (FRACUNIT/2) + (FRACUNIT/4))
		end
		
		
		if not P_IsObjectOnGround(player.mo)
			if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP)
				if player.mo.momz > 0
					player.mo.momz = $ - $ / 25
				elseif player.mo.momz < (9*FRACUNIT * player.jumpfactor)
					player.mo.momz = $ + $ / 2
				end
			else
				if player.mo.momz < 0
					player.mo.momz = $ - $ / 25
				elseif player.mo.momz < (9*FRACUNIT * player.jumpfactor)
					player.mo.momz = $ + $ / 2
				end
			end
		end
		
		if not IsPermanentEffect("gravity coil")
			player.chaos.GravityCoilTimer = $ + 1
		end
		
		if player.chaos.GravityCoilTimer >= effecttime or player.chaos.GravityCoilRemoveNow
			RemoveEffect("gravity coil")
			S_StartSound(player.mo, sfx_kc65)
			player.chaos.GravityCoilTimer = 0
			player.jumpfactor = skins[player.mo.skin].jumpfactor
			player.chaos.GravityCoilRemoveNow = false
			player.chaos.GravityCoilNow = false
		end
	end	
	
	if not player.chaos.InvertControlsNow -- random chance to invert jump/spin buttons as well?
		if CheckContiniousEffects(player, "invert controls")
			player.chaos.InvertControlsNow = true
			S_StartSound(player.mo, sfx_kc59)
			
			local RANDY = P_RandomChance(FRACUNIT/5)
			
			if RANDY
				player.chaos.InvertCam = true
			end
		end
	else
		local effecttime = invtablesearch(effectduration, "invert controls")*TICRATE

		player.chaos.InvertControlsNow = true
		
		if player.chaos.InvertCam -- code taken directly from "MirrorControls_v2.1.lua" by QuickWolfPT
			if (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_PTERABYTE or player.mo.state == S_PLAY_FIRE)
				player.mo.angle = player.drawangle
				player.chaos.OldAngle = player.mo.angle
				return
			end

			if (player.mo.state == S_PLAY_CLING or player.mo.state == S_PLAY_CLIMB or (player.onconveyor == 2 and player.mo.state == S_PLAY_PAIN) or player.powers[pw_carry] == CR_ZOOMTUBE or player.powers[pw_carry] == CR_PTERABYTE)
				player.chaos.OldAngle = player.mo.angle
				return
			end

			if ((player.pflags & PF_ANALOGMODE) and 1 or 0)
				player.chaos.OldAngle = $ or player.mo.angle

				local angdiff = (player.cmd.angleturn << FRACBITS) - player.chaos.OldAngle
				player.mo.angle = $ - 2 * angdiff

				player.chaos.OldAngle = player.mo.angle
			end
		end
		
		if not IsPermanentEffect("invert controls")
			player.chaos.InvertControlsTimer = $ + 1
		end
		
		if player.chaos.InvertControlsTimer >= effecttime or player.chaos.InvertControlsRemoveNow
			RemoveEffect("invert controls")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.InvertControlsTimer = 0
			player.chaos.InvertCam = false
			player.chaos.InvertControlsRemoveNow = false
			player.chaos.InvertControlsNow = false
		end
	end	
	
	if not player.chaos.ChangeSpriteNow -- random chance to invert jump/spin buttons as well?
		if CheckContiniousEffects(player, "change sprite")
			player.chaos.ChangeSpriteNow = true
			S_StartSound(player.mo, sfx_kc65)
			
			local RANDY = P_RandomRange(1, 6)
			
			if RANDY == 1
				--player.chaos.SpriteToChangeTo = SPR_POSS
				player.chaos.StateToChangeTo = S_POSS_RUN1
				player.chaos.SpriteMOBJ = "tank"
			elseif RANDY == 2
				player.chaos.StateToChangeTo = S_SSHELL_RUN1
				player.chaos.SpriteMOBJ = "spring shell"
			elseif RANDY == 3
				player.chaos.SpriteToChangeTo = SPR_DETN		
			elseif RANDY == 4
				player.chaos.StateToChangeTo = S_YHORIZ1
			elseif RANDY == 5
				player.chaos.StateToChangeTo = S_RING
			elseif RANDY == 6
				player.chaos.SpriteToChangeTo = SPR_EGGM

			end
			
			local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST) 
			dust.scale = FRACUNIT*2
			dust.state = S_XPLD1
		end
	else
		local effecttime = invtablesearch(effectduration, "change sprite")*TICRATE

		player.chaos.ChangeSpriteNow = true
		
		if not IsPermanentEffect("change sprite")
			player.chaos.ChangeSpriteTimer = $ + 1
		end
		
		if player.chaos.SpriteMOBJ
			if player.speed == 0 or player.skidtime > 0
				player.chaos.SpriteToChangeTo = states[player.chaos.StateToChangeTo].sprite
				player.mo.frame = 0
			else
				player.chaos.SpriteToChangeTo = nil
			end
		end
		
		if player.chaos.ChangeSpriteTimer >= effecttime or player.chaos.ChangeSpriteRemoveNow
			RemoveEffect("change sprite")
			S_StartSound(player.mo, sfx_kc65)
			player.chaos.ChangeSpriteTimer = 0
			player.chaos.ChangeSpriteRemoveNow = false
			player.chaos.ChangeSpriteNow = false
			player.chaos.SpriteToChangeTo = nil
			player.chaos.StateToChangeTo = nil
			player.chaos.SpriteMOBJ = nil
			player.chaos.SpriteMOBJ2 = nil
			player.chaos.Sprung = false
			local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST) 
			dust.scale = FRACUNIT*2
			dust.state = S_XPLD1
		end
	end
	
	if not player.chaos.MagnetNow
		if CheckContiniousEffects(player, "magnet")
			player.chaos.MagnetNow = true
			S_StartSound(player.mo, sfx_s3k41)
		end
	else
		local effecttime = invtablesearch(effectduration, "magnet")*TICRATE
		
		if not IsPermanentEffect("magnet")
			player.chaos.MagnetTimer = $ + 1
		end
		
		if chaosbad == "on"
			P_TelekinesisButBetter(player, -8*FRACUNIT, 4000*FRACUNIT)
		else
			P_TelekinesisButBetter(player, -4*FRACUNIT, 3000*FRACUNIT)
		end
		
		if player.chaos.MagnetTimer >= effecttime or player.chaos.MagnetRemoveNow
			RemoveEffect("magnet")
			S_StartSound(player.mo, sfx_kc65)
			player.chaos.MagnetTimer = 0
			player.checkedmagnet = false
			player.chaos.MagnetRemoveNow = false
			player.chaos.MagnetNow = false
		end
	end
	
	if not player.chaos.BlackHoleNow
		if CheckContiniousEffects(player, "black hole")
			player.chaos.BlackHoleNow = true
			S_StartSound(player.mo, sfx_s3k74)
		end
	else
		local effecttime = invtablesearch(effectduration, "black hole")*TICRATE
		
		if not IsPermanentEffect("black hole")
			player.chaos.BlackHoleTimer = $ + 1
		end
		
		if (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_ROLLOUT)
			debugprint("balck hole, but in minecart or rollout rock, aborting")
			player.chaos.BlackHoleRemoveNow = true
		end
		
		if player.chaos.BlackHoleTimer >= effecttime or player.chaos.BlackHoleRemoveNow
			RemoveEffect("black hole")
			S_StartSound(player.mo, sfx_kc65)
			player.chaos.BlackHoleTimer = 0
			player.checkedmagnet = false
			player.chaos.BlackHoleRemoveNow = false
			player.chaos.BlackHoleNow = false
		end
	end
	
	if not player.chaos.CharAbilityNow
		if CheckContiniousEffects(player, "no charability")
			S_StartSound(player.mo, sfx_kc59)
			player.chaos.OGdrillmeter = player.drillmeter
			player.chaos.CharAbilityNow = true
		end
	else
		local effecttime = invtablesearch(effectduration, "no charability")*TICRATE
		
		if not IsPermanentEffect("no charability")
			player.chaos.CharAbilityTimer = $ + 1
		end
		
		player.charability = CA_NONE
		player.charability2 = CA2_NONE
		player.drillmeter = 0
		player.dashmode = 0
		
		if chaosbad == "on"
			player.charflags = $ | SF_NOJUMPSPIN
			player.charflags = $ | SF_NOJUMPDAMAGE
		end
		
		if player.chaos.CharAbilityTimer >= effecttime or player.chaos.CharAbilityRemoveNow
			RemoveEffect("no charability")
			S_StartSound(player.mo, sfx_kc5a)
			player.chaos.CharAbilityTimer = 0
			player.charability = skins[player.mo.skin].ability
			player.charability2 = skins[player.mo.skin].ability2
			player.charflags = skins[player.mo.skin].flags
			player.drillmeter = player.chaos.OGdrillmeter
			player.chaos.OGdrillmeter = nil
			player.chaos.CharAbilityRemoveNow = false
			player.chaos.CharAbilityNow = false
		end
	
	end
	
	
	-- [[ Misc. Stuff ]] --
	
	if player.chaos.RestoreViewAngle
		if player.chaos.SpinOtherWay
			if player.viewrollangle < 0
				player.viewrollangle = $ + (ANG10/4)
				player.mo.rollangle = $ - (ANG10/4)
			else
				player.viewrollangle = 0
				player.mo.rollangle = 0
				player.chaos.RestoreViewAngle = false
				player.chaos.SpinOtherWay = false
			end
		else
			if player.viewrollangle > 0
				player.viewrollangle = $ - (ANG10/4)
				player.mo.rollangle = $ + (ANG10/4)
			else
				player.viewrollangle = 0
				player.mo.rollangle = 0
				player.chaos.RestoreViewAngle = false
			end
		end
	end
	
	--if player.chaos.Sprung and P_IsObjectOnGround(player)
	--	player.chaos.Sprung = false
	--end
	
	if player.chaos.SpawnTimer < 3
		player.chaos.SpawnTimer = $ + 1
	else
		player.chaos.JustSpawned = false
	end
	
	if chaosbad == "on"
		player.badmodehud = true
	else
		player.badmodehud = false
	end
	
	-- [ Remove previous effects after 18 seconds ] --
	
	if player.chaos.PreviousEffect1
		player.chaos.PE1Timer = $ + 1
		if player.chaos.PE1Timer >= player.chaos.PE1TimerExpire
			player.chaos.PE1Timer = 0
			debugprint("removed previous effect 1: " .. player.chaos.PreviousEffect1)
			player.chaos.PreviousEffect1 = nil
		end
	end
	if player.chaos.PreviousEffect2
		player.chaos.PE2Timer = $ + 1
		if player.chaos.PE2Timer >= player.chaos.PE2TimerExpire
			player.chaos.PE2Timer = 0
			debugprint("removed previous effect 2: " .. player.chaos.PreviousEffect2)
			player.chaos.PreviousEffect2 = nil
		end
	end
	if player.chaos.PreviousEffect3
		player.chaos.PE3Timer = $ + 1
		if player.chaos.PE3Timer >= player.chaos.PE3TimerExpire
			player.chaos.PE3Timer = 0
			debugprint("removed previous effect 3: " .. player.chaos.PreviousEffect3)
			player.chaos.PreviousEffect3 = nil
		end
	end
	if player.chaos.PreviousEffect4
		player.chaos.PE4Timer = $ + 1
		if player.chaos.PE4Timer >= player.chaos.PE4TimerExpire
			player.chaos.PE4Timer = 0
			debugprint("removed previous effect 4: " .. player.chaos.PreviousEffect4)
			player.chaos.PreviousEffect4 = nil
		end
	end

	-- [ HUD TEXT Change ] --
	
	--[[if player.chaos.HUDTEXTChange
		player.chaos.HUDTextChangetimer = $ + 1
		if player.chaos.HUDTextChangetimer > 3
			player.chaos.HUDTEXTChange = false
			player.chaos.HUDTEXT2 = nil
			player.chaos.HUDTextChangetimer = 0
		end
	end]]
	
	if player.chaos.ReadHelp
		player.chaos.ReadHelpTimer = $ + 1
		if player.chaos.ReadHelpTimer > 8*TICRATE
			player.chaos.ReadHelp = false
			player.chaos.ReadHelpTimer = 0
		end
	end


end)

-- [[[ ghost effect, change sprite, and randomize sfx]]] --

addHook("PostThinkFrame", function(player)
	for player in players.iterate do
		if player.chaos.GhostNow then  -- thanks to Faye on the SRB2 discord for discovering how to get MF2_SHADOW working
			player.mo.flags2 = $ | MF2_SHADOW
		end
		if player.chaos.GhostBegone
			player.mo.flags2 = $ & ~MF2_SHADOW
			player.chaos.GhostBegone = false
		end
		
		if player.chaos.FrozenNow and not player.chaos.ChangeSpriteNow and player.chaos.FrozenState
			player.mo.state = player.chaos.FrozenState
			player.mo.frame = player.chaos.FrozenFrame
			player.powers[pw_nocontrol] = 1
			if player.followmobj
				player.followmobj.state = player.chaos.FollowFrozenState
				player.followmobj.frame = player.chaos.FollowFrozenFrame
			end
		end
		
		local function RSFXRPLC(sfx)
			if sfx == nil return end
			S_StopSoundByID(player.mo, sfx)
			S_StartSound(player.mo, SoundRand())
			--S_StartSound(player.mo, P_RandomRange(1, 735)) -- 1, 735
		end
		
		if player.chaos.RandomSFXNow
			local gaysfx
			if S_SoundPlaying(player.mo, sfx_thok)
				gaysfx = sfx_thok
			elseif S_SoundPlaying(player.mo, sfx_jump)
				gaysfx = sfx_jump
			elseif S_SoundPlaying(player.mo, sfx_pudpud)
				gaysfx = sfx_pudpud
			elseif S_SoundPlaying(player.mo, sfx_putput) -- saves the zoo
				gaysfx = sfx_putput
			elseif S_SoundPlaying(player.mo, sfx_spin)
				gaysfx = sfx_spin
			elseif S_SoundPlaying(player.mo, sfx_spndsh)
				gaysfx = sfx_spndsh
			elseif S_SoundPlaying(player.mo, sfx_zoom)
				gaysfx = sfx_zoom
			elseif S_SoundPlaying(player.mo, sfx_skid)
				gaysfx = sfx_skid
			elseif S_SoundPlaying(player.mo, sfx_shield)
				gaysfx = sfx_shield
			elseif S_SoundPlaying(player.mo, sfx_wirlsg)
				gaysfx = sfx_wirlsg
			elseif S_SoundPlaying(player.mo, sfx_forcsg)
				gaysfx = sfx_forcsg
			elseif S_SoundPlaying(player.mo, sfx_elemsg)
				gaysfx = sfx_elemsg
			elseif S_SoundPlaying(player.mo, sfx_armasg)
				gaysfx = sfx_armasg			
			elseif S_SoundPlaying(player.mo, sfx_supert)
				gaysfx = sfx_supert
			elseif S_SoundPlaying(player.mo, sfx_attrsg)
				gaysfx = sfx_attrsg
			elseif S_SoundPlaying(player.mo, sfx_spkdth)
				gaysfx = sfx_spkdth
			elseif S_SoundPlaying(player.mo, sfx_splash)
				gaysfx = sfx_splash			
			elseif S_SoundPlaying(player.mo, sfx_splish)
				gaysfx = sfx_splish				
			elseif S_SoundPlaying(player.mo, sfx_strpst)
				gaysfx = sfx_strpst			
			elseif S_SoundPlaying(player.mo, sfx_kc4d)
				gaysfx = sfx_kc4d
			elseif S_SoundPlaying(player.mo, sfx_s3k7e)
				gaysfx = sfx_s3k7e			
			elseif S_SoundPlaying(player.mo, sfx_gasp)
				gaysfx = sfx_gasp		
			elseif S_SoundPlaying(player.mo, sfx_s3k45)
				gaysfx = sfx_s3k45			
			elseif S_SoundPlaying(player.mo, sfx_s3k44)
				gaysfx = sfx_s3k44		
			elseif S_SoundPlaying(player.mo, sfx_s3k43)
				gaysfx = sfx_s3k43			
			elseif S_SoundPlaying(player.mo, sfx_s3k42)
				gaysfx = sfx_s3k42				
			elseif S_SoundPlaying(player.mo, sfx_s3k40)
				gaysfx = sfx_s3k40				
			elseif S_SoundPlaying(player.mo, sfx_s3k51)
				gaysfx = sfx_s3k51			
			elseif S_SoundPlaying(player.mo, sfx_s3ka6)
				gaysfx = sfx_s3ka6		
			elseif S_SoundPlaying(player.mo, sfx_wdjump)
				gaysfx = sfx_wdjump			
			elseif S_SoundPlaying(player.mo, sfx_attrsg)
				gaysfx = sfx_attrsg
			end
			RSFXRPLC(gaysfx)
			gaysfx = nil 
		end
		
		if player.chaos.ChangeSpriteNow -- SPR_SSWY, S_YHORIZ1
			if player.chaos.SpriteToChangeTo
				player.mo.sprite = player.chaos.SpriteToChangeTo
				player.mo.frame = 0
			else
				player.mo.sprite = states[player.chaos.StateToChangeTo].sprite
				player.mo.frame = states[player.chaos.StateToChangeTo].frame
				player.chaos.StateToChangeToTimer = $ + 1
				if player.chaos.StateToChangeTo == S_RING
					if player.chaos.StateToChangeToTimer <= states[S_RING].var1
						player.mo.frame = player.chaos.StateToChangeToTimer
					else
						player.chaos.StateToChangeToTimer = 0
					end
				end
				
				
				if player.chaos.StateToChangeTo == S_POSS_RUN1
					if player.chaos.StateToChangeToTimer >= states[S_POSS_RUN1].tics
						player.chaos.StateToChangeTo = S_POSS_RUN2
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_POSS_RUN2
					if player.chaos.StateToChangeToTimer >= states[S_POSS_RUN2].tics
						player.chaos.StateToChangeTo = S_POSS_RUN3
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_POSS_RUN3
					if player.chaos.StateToChangeToTimer >= states[S_POSS_RUN3].tics
						player.chaos.StateToChangeTo = S_POSS_RUN4
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_POSS_RUN4
					if player.chaos.StateToChangeToTimer >= states[S_POSS_RUN4].tics
						player.chaos.StateToChangeTo = S_POSS_RUN5
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_POSS_RUN5
					if player.chaos.StateToChangeToTimer >= states[S_POSS_RUN5].tics
						player.chaos.StateToChangeTo = S_POSS_RUN6
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_POSS_RUN6
					if player.chaos.StateToChangeToTimer >= states[S_POSS_RUN6].tics
						player.chaos.StateToChangeTo = S_POSS_RUN1
						player.chaos.StateToChangeToTimer = 0
					end
				end
				
				if player.chaos.StateToChangeTo == S_SSHELL_RUN1
					if player.chaos.StateToChangeToTimer >= states[S_SSHELL_RUN1].tics
						player.chaos.StateToChangeTo = S_SSHELL_RUN2
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_SSHELL_RUN2
					if player.chaos.StateToChangeToTimer >= states[S_SSHELL_RUN2].tics
						player.chaos.StateToChangeTo = S_SSHELL_RUN3
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_SSHELL_RUN3
					if player.chaos.StateToChangeToTimer >= states[S_SSHELL_RUN3].tics
						player.chaos.StateToChangeTo = S_SSHELL_RUN4
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_SSHELL_RUN4
					if player.chaos.StateToChangeToTimer >= states[S_SSHELL_RUN4].tics
						player.chaos.StateToChangeTo = S_SSHELL_RUN1
						player.chaos.StateToChangeToTimer = 0
					end
				end
				
				if player.chaos.StateToChangeTo == S_YHORIZ1 and player.chaos.Sprung
					if player.chaos.StateToChangeToTimer >= states[S_YHORIZ1].tics
						player.chaos.StateToChangeTo = S_YHORIZ2
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_YHORIZ2
					if player.chaos.StateToChangeToTimer >= states[S_YHORIZ2].tics
						player.chaos.StateToChangeTo = S_YHORIZ3
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_YHORIZ3
					if player.chaos.StateToChangeToTimer >= states[S_YHORIZ3].tics
						player.chaos.StateToChangeTo = S_YHORIZ4
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_YHORIZ4
					if player.chaos.StateToChangeToTimer >= states[S_YHORIZ4].tics
						player.chaos.StateToChangeTo = S_YHORIZ5
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_YHORIZ5
					if player.chaos.StateToChangeToTimer >= states[S_YHORIZ5].tics
						player.chaos.StateToChangeTo = S_YHORIZ6
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_YHORIZ6
					if player.chaos.StateToChangeToTimer >= states[S_YHORIZ6].tics
						player.chaos.StateToChangeTo = S_YHORIZ7
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_YHORIZ7
					if player.chaos.StateToChangeToTimer >= states[S_YHORIZ7].tics
						player.chaos.StateToChangeTo = S_YHORIZ8
						player.chaos.StateToChangeToTimer = 0
					end
				elseif player.chaos.StateToChangeTo == S_YHORIZ8
					if player.chaos.StateToChangeToTimer >= states[S_YHORIZ8].tics
						player.chaos.StateToChangeTo = S_YHORIZ1
						player.chaos.StateToChangeToTimer = 0
						player.chaos.Sprung = false
					end
				end
			end
		end
	end
end)

-- [[[ inverted controls ]]] --

addHook("PreThinkFrame", do
	for player in players.iterate do
		if player.chaos.InvertControlsNow
			player.cmd.forwardmove = -($)
			player.cmd.sidemove = -($)
		end
	
		if (player.mo.eflags & MFE_SPRUNG) and player.chaos.ChangeSpriteNow
			--debugprint("sprung")
			player.chaos.Sprung = true
		end
		
	end
	
	--p.cmd.forwardmove = $*-1
	--p.cmd.sidemove = $*-1
end)


-- [[[ Player Spawn ]]] --

hud.add(darkvisionhud, "game")

addHook("PlayerSpawn",  function(player)
	if not player.mo
		return
	end
	
	if not player.chaos
		player.chaos = {}
	end
	
	if not multiplayer
		player.chaos.CEbuttonfx = CEbuttonfx
		player.chaos.CEbutton = CEbutton
	end
	
	player.chaos.HoldButton1 = false
	player.chaos.HoldButton2 = false
	player.chaos.HoldButton3 = false
	player.chaos.HoldButton4 = false
	player.chaos.HoldButton5 = false
	
	player.chaos.PizzaSpawned = false
	
	player.chaos.NukePrevShield = nil
	
	player.chaos.NoTurningAngle = nil
	
	if not player.chaos.permeffect1
		player.chaos.permeffect1 = permeffect1
		player.chaos.permeffect2 = permeffect2
	end
	
	if ReadChaosHelp == "no" and not multiplayer and RCHNagTimes < 3
		player.chaos.ReadHelpTimer = 0
		player.chaos.ReadHelp = true
		RCHNagTimes = $ + 1
		CH_SaveSettings(player)
	end
	
	player.chaos.JustSpawned = true -- used for the HUD because it's a bitch 
	player.chaos.SpawnTimer = 0
	
	if not player.chaos.OGFriction
		player.chaos.OGFriction = player.mo.friction
	end
	
	player.chaos.SuccessfulRemoval = false
	
	player.chaos.FrozenState = nil
	player.chaos.FollowFrozenState = nil	
	player.chaos.FrozenFrame = nil
	player.chaos.FollowFrozenFrame = nil
	
	player.chaos.SpriteMOBJ = nil
	player.chaos.SpriteMOBJ2 = nil
	
	player.checkpermeffect1 = false
	player.checkpermeffect2 = false
	
	player.chaos.CementShoesSfx = false
	player.chaos.CementShoesFallSfx = false
	
	player.rerollTimer = 2
	
	player.chaos.RestoreViewAngle = false
	player.chaos.SpinOtherWay = false
	player.viewrollangle = 0
	player.mo.rollangle = 0
	
	player.chaos.HUDTEXTTIMER = 0
	player.chaos.HUDTEXT = nil
	player.chaos.HUDTextChangetimer = 0
	
	player.chaos.PizzaRemoval = nil
	
	player.mo.scale = FRACUNIT
	player.mo.friction = player.chaos.OGFriction
	
	player.chaos.FirstEffect = true
	
	player.chaos.InvertCam = false
	
	player.chaos.RouletteTimer = P_RandomRange(1, TICRATE)
	
	player.chaos.ContinuousEffect1 = nil
	player.chaos.ContinuousEffect2 = nil
	player.chaos.ContinuousEffect3 = nil
	player.chaos.ContinuousEffect4 = nil
	player.chaos.StatusEffect = nil
	
	player.chaos.PreviousEffect1 = nil
	player.chaos.PreviousEffect2 = nil
	player.chaos.PreviousEffect3 = nil
	player.chaos.PreviousEffect4 = nil
	
	player.chaos.PE1Timer = 0
	player.chaos.PE2Timer = 0
	player.chaos.PE3Timer = 0
	player.chaos.PE4Timer = 0
	
	if chaosbad == "on"
		player.chaos.PE1TimerExpire = (P_RandomRange(11, 19))*TICRATE
		player.chaos.PE2TimerExpire = (P_RandomRange(11, 19))*TICRATE
		player.chaos.PE3TimerExpire = (P_RandomRange(11, 19))*TICRATE
		player.chaos.PE4TimerExpire = (P_RandomRange(11, 19))*TICRATE	
	else
		player.chaos.PE1TimerExpire = (P_RandomRange(20, 30))*TICRATE
		player.chaos.PE2TimerExpire = (P_RandomRange(20, 30))*TICRATE
		player.chaos.PE3TimerExpire = (P_RandomRange(20, 30))*TICRATE
		player.chaos.PE4TimerExpire = (P_RandomRange(20, 30))*TICRATE
	end
	
	player.chaos.timer = 0
	player.chaos.vibecheck = false
	
	-- [[ restore abilities ]] --
	player.charability = skins[player.mo.skin].ability
	player.charability2 = skins[player.mo.skin].ability2
	player.normalspeed = skins[player.mo.skin].normalspeed
	player.runspeed = skins[player.mo.skin].runspeed
	player.actionspd = skins[player.mo.skin].actionspd
	player.thrustfactor = skins[player.mo.skin].thrustfactor
	player.jumpfactor = skins[player.mo.skin].jumpfactor
	player.charflags = skins[player.mo.skin].flags
	player.camerascale = skins[player.mo.skin].camerascale
	player.mindash = skins[player.mo.skin].mindash
	player.maxdash = skins[player.mo.skin].maxdash
	
	-- [[ stop timers ]] --
	player.chaos.ButterTimer = 0
	player.chaos.GiantTimer = 0
	player.chaos.TinyTimer = 0
	player.chaos.CementTimer = 0
	player.chaos.CharAbilityTimer = 0
	player.chaos.MagnetTimer = 0
	player.chaos.DriftingTimer = 0
	player.chaos.SkinColorTimer = 0
	player.chaos.SkinTimer = 0
	player.chaos.JumpTimer = 0
	player.chaos.SpinTimer = 0
	player.chaos.PizzaTimer = 0
	player.chaos.MoneyTimer = 0
	player.chaos.GhostTimer = 0
	player.chaos.QuakeTimer = 0
	player.chaos.GravityCoilTimer = 0
	player.chaos.InvertControlsTimer = 0
	player.chaos.ChangeSpriteTimer = 0
	player.chaos.StateToChangeToTimer = 0
	player.chaos.FunkyCameraTimer = 0
	player.chaos.LavaTimer = 0
	player.chaos.NegativeRingsTimer = 0
	player.chaos.DarkenedVisionTimer = 0
	player.chaos.NoTurningTimer = 0
	player.chaos.FrozenTimer = 0
	player.chaos.RandomSFXTimer = 0
	player.chaos.BlackHoleTimer = 0
	
	player.chaos.OneTimeFxTimer = 4*TICRATE
	
	-- [[ remove me, wait no not literally ]] --
	player.chaos.GiantRemoveNow = false
	player.chaos.CementRemoveNow = false
	player.chaos.TinyRemoveNow = false
	player.chaos.CharAbilityRemoveNow = false
	player.chaos.MagnetRemoveNow = false
	player.chaos.DriftingRemoveNow = false
	player.chaos.ButterRemoveNow = false
	player.chaos.SkinColorRemoveNow = false
	player.chaos.SkinRemoveNow = false
	player.chaos.JumpRemoveNow = false
	player.chaos.SpinRemoveNow = false
	player.chaos.PizzaRemoveNow = false		
	player.chaos.MoneyRemoveNow = false
	player.chaos.GhostRemoveNow = false
	player.chaos.QuakeRemoveNow = false
	player.chaos.GravityCoilRemoveNow = false
	player.chaos.InvertControlsRemoveNow = false
	player.chaos.ChangeSpriteRemoveNow = false
	player.chaos.FunkyCameraRemoveNow = false
	player.chaos.LavaRemoveNow = false
	player.chaos.NegativeRingsRemoveNow = false
	player.chaos.DarkenedVisionRemoveNow = false
	player.chaos.NoTurningRemoveNow = false
	player.chaos.FrozenRemoveNow = false
	player.chaos.RandomSFXRemoveNow = false
	player.chaos.BlackHoleRemoveNow = false
	
	
	-- [[ the now ]] --
	player.chaos.DriftingNow = false
	player.chaos.QuakeNow = false
	player.chaos.MoneyBagNow = false
	
	if player.chaos.GhostNow
		player.chaos.GhostNow = false
		player.chaos.GhostBegone = true
		player.pflags = $ & ~PF_INVIS
	end
	
	player.chaos.PizzaNow = false
	player.chaos.SkinColorNow = false
	player.chaos.SkinNow = false
	player.chaos.SpinNow = false
	player.chaos.ButterNow = false
	player.chaos.GiantNow = false
	player.chaos.TinyNow = false
	player.chaos.CementNow = false
	player.chaos.SuperJumpNow = false
	player.chaos.MagnetNow = false
	player.chaos.CharAbilityNow = false
	player.chaos.GravityCoilNow = false
	player.chaos.InvertControlsNow = false
	player.chaos.ChangeSpriteNow = false
	player.chaos.FunkyCameraNow = false
	player.chaos.LavaNow = false
	player.chaos.NegativeRingsNow = false
	player.chaos.DarkenedVisionNow = false
	player.chaos.NoTurningNow = false
	player.chaos.FrozenNow = false
	player.chaos.RandomSFXNow = false
	player.chaos.BlackHoleNow = false
	
	player.chaos.PizzaRemovalTimer = 0

	player.chaos.Sprung = false

end)

-- [[ map change ]] --

addHook("MapChange", function(map)
	someoneusingmagnet = false
	someoneusingblackhole = false
	mobjcount = 0
end)

addHook("MobjSpawn", function(mobj)
	mobjcount = $ + 1
end)

addHook("MobjRemoved", function(mobj)
	mobjcount = $ - 1
end)

-- [[ anti-ring and x2 ring ]] --

local function NegativeRings(ring, plyr) local player = plyr.player

	if player.chaos.NegativeRingsNow
		if chaosbad == "on"
			P_GivePlayerRings(player, -2)
		else
			P_GivePlayerRings(player, -1)
		end
		P_KillMobj(ring)
		if ring.type == MT_RING
			S_StartSound(ring, sfx_antiri)
		else
			S_StartSound(ring, sfx_antico)
		end
		ring.shadowscale = 0
		return true
	elseif player.chaos.Rings2xNow
		if chaosbad == "on"
			P_GivePlayerRings(player, 3)
		else
			P_GivePlayerRings(player, 1)
		end
		return false
	elseif player.chaos.RandomSFXNow -- sad thing is that this is netplay safe, but the rest of the code for random sfx isn't :(
		P_GivePlayerRings(player, 1)
		P_KillMobj(ring)
		S_StopSound(ring)
		S_StartSound(ring, SoundRand())
		ring.shadowscale = 0
		return true
	else
		return false
	end
	


end

addHook("TouchSpecial", NegativeRings, MT_RING)
addHook("TouchSpecial", NegativeRings, MT_COIN)

-- [[ black hole code, and vibe check ]] --

local function PSearch(mobj)
	for player in players.iterate do
		if player.chaos.BlackHoleNow and mobj.player != player
			if R_PointToDist2(mobj.x, mobj.y, player.mo.x, player.mo.y) < 1024*FRACUNIT
				return player.mo
			elseif (R_PointToDist2(mobj.x, mobj.y, player.mo.x, player.mo.y) < 1200*FRACUNIT and (mobj.flags & (MF_NOGRAVITY))) -- object is outside of the radius, and has no gravity, set everything to 0
				mobj.momx = 0
				mobj.momy = 0
				mobj.momz = 0
			end
		end
	end
end

local function Search(mobj)
	if not mobj.valid return end 
	if mobj.player return end
	if (mobj.flags & (MF_SCENERY|MF_SOLID)) == (MF_SCENERY|MF_SOLID) and not (mobj.type != MT_SPIKE and chaosbad == "on") return end -- exception for stuff like GFZ trees, too annoying to deal with
	if (mobj.flags & MF_SPRING) and chaosbad == "off" return end
	if (mobj.flags & MF_PUSHABLE) return end -- just because those gargoyles are annoying too 
	if (mobj.type == MT_EGGTRAP or mobj.type == MT_STEAM or mobj.type == MT_CHAINPOINT or mobj.type == MT_MINECART or mobj.type == MT_MINECARTSEG or mobj.type == MT_MINECARTSPAWNER or mobj.type == MT_MINECARTEND or mobj.type == MT_MINECARTENDSOLID or mobj.type == MT_MINECARTSIDEMARK or mobj.type == MT_MINECARTSPARK) return end
	
	local Searcher = PSearch(mobj)
	
	if Searcher
		if mobj.type == MT_RING or mobj.type == MT_COIN or mobj.type == MT_BLUESPHERE
			mobj.tracer = Searcher
		end
		mobj.BHTarget = Searcher
		return true
	end
end

addHook("ThinkFrame", function(thonk) 
	someoneusingblackhole = false
	someoneusingmagnet = false
	for player in players.iterate do
		if player.chaos.BlackHoleNow
			someoneusingblackhole = true
		end
		if player.chaos.MagnetNow
			someoneusingmagnet = true
		end
		
		
		-- [[ the vibes ]] --
		if player.chaos.vibecheck and not SlotsFull(player) and player.rerollTimer <= 0
		
			player.chaos.StatusRand = P_RandomRange(1, numberofeffects)
			if P_RandomChance(vibechance) or player.reroll or player.chaos.FirstEffect
				player.chaos.FirstEffect = false
				if player.reroll
					debugconsprintf(player, "*\x85\ reroll\x80")
					player.chaos.timer = 0
					player.reroll = false
				end
				player.chaos.StatusEffect = tablesearch(statuseffects, player.chaos.StatusRand)
				debugconsprintf(player, "*\x82\ rolled\x80: " .. player.chaos.StatusEffect)
				
				-- [ effect chances ] --
				if player.chaos.StatusEffect -- sfx_s3kd2s
					if P_RandomChance(invtablesearch(effectchance, player.chaos.StatusEffect)) and not CheckContiniousEffects(player, player.chaos.StatusEffect) and LastCheck(player, player.chaos.StatusEffect)
					--and player.chaos.RouletteTimer <= 0
						--[[if (player.bot == 1 and P_RandomChance(FRACUNIT/3) and not multiplayer) -- reduced chance for the bot player in Sonic & Tails mode
						or player.bot != 1]]
						ApplyEffect(player, player.chaos.StatusEffect)
						--end
					else
						player.reroll = true
						player.rerollTimer = 2
						return
					end
				end
			end
			player.chaos.vibecheck = false
		elseif SlotsFull(player)
			player.chaos.vibecheck = false
		end
	end
	
	if someoneusingblackhole
		local lagcounter = (mobjcount/800)
		if netgame
			 lagcounter = (mobjcount/590)
		end
		if someoneusingmagnet
			lagcounter = $ + 1
		end
		if lagcounter == 0
			lagcounter = 1
		end
		
		if (leveltime % lagcounter) == 0
			for mobj in mobjs.iterate() do
				if Search(mobj)
					BootlegAttract(mobj, mobj.BHTarget, 16*FRACUNIT)
				end
			end
		end
	end
end)


-- [[ the pizza is agressive ]] --

addHook("MobjThinker", function(aggressive) -- this code is directly lifted from my MMT script, i have no shame
	if aggressive.valid and aggressive.target and aggressive.target.valid
		if not aggressive.initializedtime
			aggressive.shadowscale = 3*FRACUNIT
			if aggressive.sizeoption == "large"
				aggressive.speed = 44*FRACUNIT
				aggressive.scale = FRACUNIT + FRACUNIT*(2/4)
			elseif aggressive.sizeoption == "medium"
				aggressive.speed = 39*FRACUNIT
				aggressive.scale = FRACUNIT
			elseif (aggressive.sizeoption == "small" or aggressive.sizeoption == "personal")
				aggressive.speed = (28*3)*FRACUNIT
				aggressive.scale = FRACUNIT/4
			elseif (aggressive.sizeoption == "xl" or aggressive.sizeoption == "extra large" or aggressive.sizeoption == "extralarge")
				aggressive.speed = 49*FRACUNIT
				aggressive.scale = FRACUNIT*2
			else
				aggressive.speed = (36*2)*FRACUNIT
				aggressive.scale = FRACUNIT/2
			end
			aggressive.initializedtime = true
		end
		if not aggressive.rollanglesprite
		aggressive.rollanglesprite = 0
		end
		aggressive.rollanglesprite = $ + ANG15
		aggressive.rollangle = aggressive.rollanglesprite

		if (R_PointToDist2(aggressive.target.x, aggressive.target.y, aggressive.x, aggressive.y))/FRACUNIT < 35 and (abs((aggressive.target.z - aggressive.z)/FRACUNIT) < 25) and not (aggressive.target.player and aggressive.target.player.playerstate == PST_DEAD) and not aggressive.Remove
			if chaosbad == "on"
				P_RemoveShield(aggressive.target.player)
				aggressive.target.player.powers[pw_invulnerability] = 0
				P_DamageMobj(aggressive.target, aggressive, aggressive, DMG_SPIKE)
			else
				P_DamageMobj(aggressive.target, aggressive, aggressive, DMG_SPIKE)
			end
			
			S_StartSound(aggressive.target, sfx_chapki)
			if aggressive.state == S_CHAOSEXPLOSPIZZA
				local Explode = P_SpawnMobj(aggressive.target.x, aggressive.target.y, aggressive.target.z, MT_BOSSEXPLODE)
				S_StartSound(explode, sfx_bkpoof)
				Explode.destscale = 3*FRACUNIT
				Explode.scalespeed = FRACUNIT/7
			end
			aggressive.target.player.chaos.PizzaRemoveNow = true
			aggressive.target.player.chaos.PizzaHitTarget = true
			aggressive.Remove = true
		end

		if aggressive.target and not aggressive.target.player.chaos.PizzaRemoval
			A_HomingChase(aggressive, aggressive.speed, 0)
		else
			P_RemoveMobj(aggressive)
		end
	else
		P_RemoveMobj(aggressive)
	end
end, MT_JIMBO)

addHook("GameQuit", function(gamer) -- doesn't work if exiting the game from the title screen, i don't know why

	if not multiplayer
		CH_SaveSettings()
	end

end)

local chaossettingsexist

if not multiplayer 
	chaossettingsexist = io.openlocal("chaosedition/chaosedition.txt", "r") -- yeah i basically stole all this from mmt. what? it's my own code,
end

--if io.openlocal("chaosedition/chaosedition.txt", "r") == chaosversion chaossettingsexist = true end

if not multiplayer
	-- create files if they don't already exist
	if chaossettingsexist
		debugprint("yes chaossettingsexist")
	else
		debugprint("no chaossettingsexist or wrong version, creating files")
		local CHversion = io.openlocal("chaosedition/chaosedition.txt", "w")
		CHversion:write(chaosversion)
		CHversion:close()
		CH_SaveSettings()
	end
	
	-- autoloading
	
	local TimeToLoad = io.openlocal("chaosedition/chaosautoload.txt", "r")
	if TimeToLoad and TimeToLoad:read("*all") == "on"
		debugprint("loading in settings")
		CH_LoadSettings()
	else
		debugprint("not loading in settings")
	end
	
	if TimeToLoad
		TimeToLoad:close()
	end
	
end

addHook("NetVars", function(net)
	vibechance = net(vibechance)
	chaosdebugger = net(chaosdebugger)
	statuseffects = net(statuseffects)
	someoneusingmagnet = net(someoneusingmagnet)
	chaosbad = net(chaosbad)
	minwaittime = net(minwaittime)
	maxwaittime = net(maxwaittime)
	mobjcount = net(mobjcount)
	bannedeffects = net(bannedeffects)
end)

print("\x81\SRB2 Chaos Edition \x8B" .. chaosversion .. "\x80 Loaded successfully!")
if ReadChaosHelp == "no" print("Please pause & open the console (press\x87 ~\x80) and type\x87 chaoshelp\x80!") end