--[[ chaos edition hud

GOALS:
	- Make hud element that properly displays what status effects are afflicted on the player. There are 4 in total, using player.chaos.ContinuousEffect1 as the thing.
	- When the script is choosing an effect, it will rapidly cycle through the effects in player.chaos.StatusEffect, until it eventually lands on a good roll and chooses that effect.

INFO:
	- Hud elements are rendered from the top-left most pixel.
	- We are aiming for bottom center of screen. Ergo, calculate from lowest resolution, account for Y-postion in all resolutions, go from there.
	
	- 3F3F3F = Neutral/No Effect
	- 6F0000 = Negative Effect
	- 005F00 = Positive Effect

TODO:
	- [ ] Add player.badmodehud, switch hud texture to CEIBC0
]]

freeslot(
"SPR_CEIB",
"SPR_CEIC" -- A0 is blank, use other letters for more stuff
)

local function invtablesearch(tbl, val)
    for k, v in pairs(tbl) do
        if k == val then 
			return v
		end
    end
    return nil
end

local function chaos(v, player)

	if player.chaos.chaoshud == "off"
		return
	end
		
	if player.chaos.chaoshud == "bottom"
		if player.chaos.chaoshudpos == 0
			v.offsety = -15
		elseif player.chaos.chaoshudpos == 1
			v.offsety = -5
		elseif player.chaos.chaoshudpos == 2
			v.offsety = 10
		elseif player.chaos.chaoshudpos == 3
			v.offsety = 28
		elseif player.chaos.chaoshudpos == 4
			v.offsety = 38
		elseif player.chaos.chaoshudpos == 5
			v.offsety = 50
		end
	elseif player.chaos.chaoshud == "top"
		if player.chaos.chaoshudpos == 0
			v.offsety = -163 - -15
		elseif player.chaos.chaoshudpos == 1
			v.offsety = -163 - -5
		elseif player.chaos.chaoshudpos == 2
			v.offsety = -163 - 10
		elseif player.chaos.chaoshudpos == 3
			v.offsety = -163 - 28
		elseif player.chaos.chaoshudpos == 4
			v.offsety = -163 - 38		
		elseif player.chaos.chaoshudpos == 5
			v.offsety = -163 - 50
		end
	end
	
	if player.badmodehud
		v.HUDBorderTexture = "CEIBC0"
	else
		v.HUDBorderTexture = "CEIBA0"
	end

	if player.chaos.permeffect1 != "none" v.drawScaled(87*FRACUNIT, (171+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch("CEIBB0"), V_PERPLAYER) end
	v.drawScaled(112*FRACUNIT, (171+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.HUDBorderTexture), V_PERPLAYER) -- offsets are 25
	v.drawScaled(137*FRACUNIT, (171+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.HUDBorderTexture), V_PERPLAYER)
	v.drawScaled(162*FRACUNIT, (171+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.HUDBorderTexture), V_PERPLAYER)
	v.drawScaled(187*FRACUNIT, (171+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.HUDBorderTexture), V_PERPLAYER)
	if player.chaos.permeffect2 != "none" v.drawScaled(212*FRACUNIT, (171+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch("CEIBB0"), V_PERPLAYER) end
	
	-- [[ the effects ]] --
	
	local function ApplyEffect(CENUM, timer, sprite)
		if timer < (invtablesearch(chaosEffectDuration, CENUM)-(v.FLASHTIME))*TICRATE
			return sprite
		elseif (timer % 2 == 0) -- if timer is divisible by 2
				return sprite
			else
				if player.badmodehud
					return "CEICW0"
				else
					return "CEICA0"
				end
			end
	end
	
	
	local function EffectSprite(CENUM, timez)
		
		if CENUM == 1
			CENUM = player.chaos.ContinuousEffect1
		elseif CENUM == 2
			CENUM = player.chaos.ContinuousEffect2
		elseif CENUM == 3
			CENUM = player.chaos.ContinuousEffect3
		elseif CENUM == 4
			CENUM = player.chaos.ContinuousEffect4
		elseif CENUM == 5
			CENUM = player.chaos.StatusEffect
		elseif CENUM == 6
			CENUM = player.chaos.permeffect1
		elseif CENUM == 7
			CENUM = player.chaos.permeffect2
		else
			error("Number provided for EffectSprite is invalid!", 1)
		end
		
		v.FLASHTIME = 2
		
		if CENUM == "giant"
			if timez
				return abs((player.chaos.GiantTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.GiantTimer, "CEICC0")
			end
		
		-- [ timer fx ] --
		elseif CENUM == "cement shoes"
			if timez
				return abs((player.chaos.CementTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.CementTimer, "CEICD0")
			end
		elseif CENUM == "no charability"
			if timez
				return abs((player.chaos.CharAbilityTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.CharAbilityTimer, "CEICE0")
			end
		elseif CENUM == "magnet"
			if timez
				return abs((player.chaos.MagnetTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.MagnetTimer, "CEICF0")
			end
		elseif CENUM == "black hole"
			if timez
				return abs((player.chaos.BlackHoleTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.BlackHoleTimer, "CEIC50")
			end
		elseif CENUM == "tiny"
			if timez
				return abs((player.chaos.TinyTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.TinyTimer, "CEICG0")
			end
		elseif CENUM == "drifting left"
			if timez
				return abs((player.chaos.DriftingTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.DriftingTimer, "CEICH0")	
			end
		elseif CENUM == "butter physics"
			if timez
				return abs((player.chaos.ButterTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.ButterTimer, "CEICI0")
			end
		elseif CENUM == "change skin color"
			if timez
				return abs((player.chaos.SkinColorTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.SkinColorTimer, "CEICJ0")
			end
		elseif CENUM == "change skin"
			if timez
				return abs((player.chaos.SkinTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.SkinTimer, "CEICK0")
			end
		elseif CENUM == "super jump"
			if timez
				return abs((player.chaos.JumpTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.JumpTimer, "CEICL0")
			end
		elseif CENUM == "the ride never ends" 
			if timez
				return abs((player.chaos.SpinTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.SpinTimer, "CEICB0") -- fun fact: CEICB0 is a place holder sprite, but i decided to use it here since it's fitting		
			end
		elseif CENUM == "the pizza is aggressive" 
			if timez
				return abs((player.chaos.PizzaTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.PizzaTimer, "CEICM0")
			end
		elseif CENUM == "money bag" 
			if timez
				return abs((player.chaos.MoneyTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.MoneyTimer, "CEICN0")
			end
		elseif CENUM == "ghost" 
			if timez
				return abs((player.chaos.GhostTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.GhostTimer, "CEICT0")
			end
		elseif CENUM == "earthquake" 
			if timez
				return abs((player.chaos.QuakeTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.QuakeTimer, "CEICU0")
			end
		elseif CENUM == "gravity coil" 
			if timez
				return abs((player.chaos.GravityCoilTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.GravityCoilTimer, "CEICV0")
			end
		elseif CENUM == "invert controls" 
			if timez
				return abs((player.chaos.InvertControlsTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.InvertControlsTimer, "CEICX0")
			end
		elseif CENUM == "change sprite" 
			if timez
				return abs((player.chaos.ChangeSpriteTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.ChangeSpriteTimer, "CEICY0")
			end
		elseif CENUM == "funky camera" 
			if timez
				return abs((player.chaos.FunkyCameraTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.FunkyCameraTimer, "CEICZ0")
			end
		elseif CENUM == "the floor is lava" 
			if timez
				return abs((player.chaos.LavaTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.LavaTimer, "CEIC10")
			end
		elseif CENUM == "negative rings" 
			if timez
				return abs((player.chaos.NegativeRingsTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.NegativeRingsTimer, "CEIC20")
			end
		elseif CENUM == "darkened vision" 
			if timez
				return abs((player.chaos.DarkenedVisionTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.DarkenedVisionTimer, "CEIC30")
			end
		elseif CENUM == "no turning" 
			if timez
				return abs((player.chaos.NoTurningTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.NoTurningTimer, "CEIC40")
			end
		elseif CENUM == "i was frozen today" 
			if timez
				return abs((player.chaos.FrozenTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.FrozenTimer, "CEIC60")
			end
		elseif CENUM == "randomize sfx" 
			if timez
				return abs((player.chaos.RandomSFXTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.RandomSFXTimer, "CEIC70") -- make this a speaker icon, like the one from windows xp or 98
			end

		-- [ one-time fx ] --
		elseif CENUM == "remove random effect" 
			if timez
				return abs((player.chaos.OneTimeFxTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.OneTimeFxTimer, "CEICO0")		
			end
		elseif CENUM == "superring" 
			if timez
				return abs((player.chaos.OneTimeFxTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.OneTimeFxTimer, "CEICP0")		
			end
		elseif CENUM == "anti-superring" 
			if timez
				return abs((player.chaos.OneTimeFxTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.OneTimeFxTimer, "CEICQ0")		
			end
		elseif CENUM == "i taketh" 
			if timez
				return abs((player.chaos.OneTimeFxTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.OneTimeFxTimer, "CEICR0")		
			end
		elseif CENUM == "nuke" 
			if timez
				return abs((player.chaos.OneTimeFxTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.OneTimeFxTimer, "CEICS0")
			end
		elseif CENUM == "random music" 
			if timez
				return abs((player.chaos.OneTimeFxTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.OneTimeFxTimer, "CEIC00")
			end
		elseif CENUM == "spawn object"
			if timez
				return abs((player.chaos.OneTimeFxTimer/TICRATE - invtablesearch(chaosEffectDuration, CENUM)))
			else
				return ApplyEffect(CENUM, player.chaos.OneTimeFxTimer, "CEIC80")
			end

		else
			if timez
				return "0"
			else
				return "CEICB0"
			end
		end

	end
	
	v.Section1TEXT = "0"
	v.Section2TEXT = "0"
	v.Section3TEXT = "0"
	v.Section4TEXT = "0"
	
	-- CEICW0
	
	if player.chaos.ContinuousEffect1
		v.Section1 = EffectSprite(1)
		v.Section1TEXT = tostring(EffectSprite(1, true))
	else
		if player.badmodehud
			v.Section1 = "CEICW0"
		else
			v.Section1 = "CEICA0"
		end
	end
	
	if player.chaos.ContinuousEffect2
		v.Section2 = EffectSprite(2)
		v.Section2TEXT = tostring(EffectSprite(2, true))
	else
		if player.badmodehud
			v.Section2 = "CEICW0"
		else
			v.Section2 = "CEICA0"
		end
	end
	
	if player.chaos.ContinuousEffect3
		v.Section3 = EffectSprite(3)
		v.Section3TEXT = tostring(EffectSprite(3, true))
	else
		if player.badmodehud
			v.Section3 = "CEICW0"
		else
			v.Section3 = "CEICA0"
		end
	end
	
	if player.chaos.ContinuousEffect4
		v.Section4 = EffectSprite(4)
		v.Section4TEXT = tostring(EffectSprite(4, true))
	else
		if player.badmodehud
			v.Section4 = "CEICW0"
		else
			v.Section4 = "CEICA0"
		end
	end

	if player.chaos.permeffect1 != "none"
		v.Section6 = EffectSprite(6)
	else
		if player.badmodehud
			v.Section6 = "CEICW0"
		else
			v.Section6 = "CEICA0"
		end
	end
	
	if player.chaos.permeffect2 != "none" 
		v.Section7 = EffectSprite(7)
	else
		if player.badmodehud
			v.Section7 = "CEICW0"
		else
			v.Section7 = "CEICA0"
		end
	end

	if player.chaos.StatusEffect
		if not player.chaos.ContinuousEffect1
			v.Section1 = EffectSprite(5)
		elseif not player.chaos.ContinuousEffect2
			v.Section2 = EffectSprite(5)
		elseif not player.chaos.ContinuousEffect3
			v.Section3 = EffectSprite(5)
		elseif not player.chaos.ContinuousEffect4
			v.Section4 = EffectSprite(5)
		end
	end
	
	if player.chaos.HUDTEXT
		if not v.HUDTEXTtrans
			v.HUDTEXTtrans = 0
			v.HUDTEXT = player.chaos.HUDTEXT
			--print("HUDTEXT: " .. v.HUDTEXT)
		end
		
		if player.chaos.HUDText and v.HUDTEXT != player.chaos.HUDTEXT
			v.HUDTEXT = player.chaos.HUDTEXT
		end
		
		--[[if player.chaos.HUDTEXTChange
			v.HUDTEXT = player.chaos.HUDTEXT2
		end]]
		
		if player.chaos.HUDTEXTTIMER < 2*TICRATE
			v.HUDTEXTFlags = V_PERPLAYER|V_ALLOWLOWERCASE
		else
			if (((player.chaos.HUDTEXTTIMER-2*TICRATE)*2) % 5 == 0)
				v.HUDTEXTtrans = $ + 1
			end
			v.HUDTEXTFlags = V_PERPLAYER|V_ALLOWLOWERCASE|v.HUDTEXTtrans<<V_ALPHASHIFT
		end

		if player.chaos.chaoshud == "top"
			v.TEXToffsety = v.offsety + 35
		else
			v.TEXToffsety = v.offsety
		end
		
		
		if v.HUDTEXT v.drawString(114, 160+v.TEXToffsety, v.HUDTEXT, v.HUDTEXTFlags, "thin") end
	else
		v.HUDTEXT = nil
		v.HUDTEXTtrans = nil 
	end
	
	if player.chaos.JustSpawned
		v.HUDTEXTtrans = nil
		v.HUDTEXT = nil
		v.HUDTEXTFlags = nil
	end
	
	if player.chaos.permeffect1 != "none" v.drawScaled(89*FRACUNIT, (173+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.Section6), V_PERPLAYER) end
	v.drawScaled(114*FRACUNIT, (173+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.Section1), V_PERPLAYER)
	v.drawScaled(139*FRACUNIT, (173+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.Section2), V_PERPLAYER) --offset by 2
	v.drawScaled(164*FRACUNIT, (173+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.Section3), V_PERPLAYER)
	v.drawScaled(189*FRACUNIT, (173+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.Section4), V_PERPLAYER)
	if player.chaos.permeffect2 != "none" v.drawScaled(214*FRACUNIT, (173+v.offsety)*FRACUNIT, FRACUNIT, v.cachePatch(v.Section7), V_PERPLAYER) end

	if v.Section1TEXT != "0" v.drawString(114+1, 173+1+v.offsety, v.Section1TEXT, V_PERPLAYER|V_ALLOWLOWERCASE|V_30TRANS, "thin") end
	if v.Section2TEXT != "0" v.drawString(139+1, 173+1+v.offsety, v.Section2TEXT, V_PERPLAYER|V_ALLOWLOWERCASE|V_30TRANS, "thin") end
	if v.Section3TEXT != "0" v.drawString(164+1, 173+1+v.offsety, v.Section3TEXT, V_PERPLAYER|V_ALLOWLOWERCASE|V_30TRANS, "thin") end
	if v.Section4TEXT != "0" v.drawString(189+1, 173+1+v.offsety, v.Section4TEXT, V_PERPLAYER|V_ALLOWLOWERCASE|V_30TRANS, "thin") end
	
	
	-- 68, 103
	
	if player.chaos.ReadHelp
		v.drawString(158, 103, "Welcome to SRB2 Chaos Edition!", V_ALLOWLOWERCASE|V_30TRANS, "thin-center")
		v.drawString(158, 113, "Please pause & open the console (press\x87 ~\x80) and type\x87 chaoshelp\x80!", V_ALLOWLOWERCASE|V_30TRANS, "thin-center")
	end

end

hud.add(chaos, "game")